from .counterpoint import create_counterpoint, create_counterpoint_on_chord, create_counterpoint_on_score
from .project import project_on_one_chord, project_on_rhythm, project_on_score_keep_notes, project_on_several_chord
from .voice_leading import VoiceLeading, Rules
from .patternator import Patternator
import musiclang.transform.composing.pattern as pattern
from .layer import MelodicLayer, OrchestralLayer, PartComposer, GlobalLayer


"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""
from musiclang.library import *
import re

class Score:
    """Represents a score in musiclang


    """
    def __init__(self, chords=None, config=None, tags=None):
        self.chords = chords
        self.config = config
        if self.chords is None:
            self.chords = []
        if self.config is None:
            self.config = {'annotation': "",
                           "tempo": 120,
                           "pickup": 0,
                           "time_signature": (4, 4)
                           }

        self.tags = set(tags) if tags is not None else set()

    def has_tag(self, tag):
        """
        Check if the tag exists for this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        chord: Chord
        """
        return tag in self.tags

    def add_tag(self, tag):
        """
        Add a tag to this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        chord: Chord
        """
        cp = self.copy()
        cp.tags.add(tag)
        return cp

    def add_tags(self, tags):
        """
        Add several tags to the object.
        Returns a copy of the object

        Parameters
        ----------
        tags: List[str]
        tags to add

        Returns
        -------
        score: Score

        """
        cp = self.copy()
        cp.tags = cp.tags.union(set(tags))
        return cp

    def add_tag_children(self, tag):
        """
        Add a tag to each chord composing the score (A score itself don't have tags)
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        chord: Chord
        """
        return Score([c.add_tag(tag) for c in self.chords], tags=self.tags)


    def add_tags_children(self, tags):
        """
        Add several tags to each chord composing the score (A score itself don't have tags)
        Returns a copy of the object

        Parameters
        ----------
        tags: List[str]
        tags to add

        Returns
        -------
        score: Score

        """
        return Score([c.add_tags(tags) for c in self.chords], tags=self.tags)

    def remove_tags_children(self, tags):
        """
        Remove several tags to each chord composing the score (A score itself don't have tags)
        Returns a copy of the object

        Parameters
        ----------
        tags: List[str]
        tags to add

        Returns
        -------
        score: Score

        """
        return Score([c.remove_tags(tags) for c in self.chords], tags=self.tags)

    def remove_tag_children(self, tag):
        """
        Remove a tag to each chord composing the score (A score itself don't have tags)
        Returns a copy of the object

        Parameters
        ----------
        tags: List[str]
        tags to add

        Returns
        -------
        score: Score

        """
        return Score([c.remove_tag(tag) for c in self.chords], tags=self.tags)

    def clear_tags_children(self):
        """
        Clear all tags to each chord composing the score (A score itself don't have tags)
        Returns a copy of the object

        Parameters
        ----------
        tags: List[str]
        tags to add

        Returns
        -------
        score: Score

        """
        return Score([c.clear_tags() for c in self.chords], tags=self.tags)

    def remove_tags(self, tags):
        """
        Remove several tags from the object.
        Returns a copy of the object

        Parameters
        ----------
        tags: List[str]

        Returns
        -------
        score: Score


        """
        cp = self.copy()
        cp.tags = cp.tags - set(tags)
        return cp

    def set_duration(self, duration):
        return Score([c.set_duration(duration) for c in self.chords], tags=set(self.tags))

    def remove_tag(self, tag):
        """
        Remove a tag from this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        chord: Chord
        """
        cp = self.copy()
        cp.tags.remove(tag)
        return cp

    def clear_tags(self):
        """
        Clear all tags from this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        chord: Chord
        """
        cp = self.copy()
        cp.tags = set()
        return cp


    def set_amp(self, amp):
        """
        Set the velocity of the notes in the score (between 0 and 120)
        Parameters
        ----------
        amp: int

        Returns
        -------

        """
        return Score([c.set_amp(amp) for c in self.chords], tags=set(self.tags))


    def normalize_instruments(self):
        """
        Get a score with the same instruments on all chords,
        if an instrument is missing on one chord add it with a silence lasting for the chord duration.

        Returns
        -------
        score: Score
        Normalize score
        """
        from musiclang import Silence
        score = []
        instruments = set(self.instruments)
        for chord_raw in self.chords:
            chord = chord_raw.copy()
            missing_instruments = set(instruments) - set(chord.instruments)
            chord_score = chord.score
            for instrument in missing_instruments:
                chord_score[instrument] = Silence(chord.duration)
            score.append(chord(**chord_score))
        return Score(score)

    def remove_silenced_instruments(self):
        """
        Inverse operation of :func:`~Score.normalize_instruments`
        For each chord remove all instruments that are silenced
        Returns
        -------
        score: Score
        """

        chords = []
        for chord_raw in self.chords:
            chord = chord_raw.copy()
            chord_score = chord.score

            for instrument in list(chord_score.keys()):
                if all([n.type == 'r' for n in chord_score[instrument]]):
                    del chord_score[instrument]

            chords.append(chord(**chord_score))

        return sum(chords, None)

    def normalize_instrument_names(self):
        """
        Normalize the instrument idx for all instrument.
        The first voice of an instrument encountered will always have idx 0, the second idx 1, etc...
        Returns
        -------
        score: Score
        """

        instruments = self.instruments

        replace_dic = {}
        replace_dic_ins = {}
        for ins in instruments:
            name, idx = '__'.join(ins.split('__')[:-1]), ins.split('__')[-1]
            replace_dic_ins[name] = replace_dic_ins.get(name, -1) + 1
            replace_dic[ins] = name + '__' + str(replace_dic_ins[name])

        score = self.replace_instruments(**replace_dic)
        return score

    def replace_instruments(self, **instruments_dict):
        """
        Replace any instrument with another (use full part name (eg: piano__0)

        Parameters
        ----------
        instruments_dict: dict[str, str]
            Dictionary of parts name to replace

        Returns
        -------
        score: Score

        """
        score = []
        instruments_dict = {key: instruments_dict[key] if key in instruments_dict.keys() else key for key in self.parts}
        for chord in self.chords:
            new_chord_dict = {}
            for ins_name, new_ins_name in instruments_dict.items():
                if ins_name in chord.score.keys():
                    new_chord_dict[new_ins_name] = chord.score[ins_name]
            score.append(chord(**new_chord_dict))

        return Score(score)


    def replace_instruments_names(self, **instruments_dict):
        """
        Replace any instrument name with another (use only instrument name (eg: piano)

        Parameters
        ----------
        instruments_dict: dict[str, str]
            Dictionary of parts name to replace

        Returns
        -------
        score: Score

        """
        score = []
        for chord in self.chords:
            score.append(chord.replace_instruments_names(**instruments_dict))

        return Score(score)


    def clean(self):
        return self.to_drum().decompose_duration()

    def to_drum(self):
        return Score([chord.to_drum() for chord in self.chords], tags=set(self.tags))

    def to_chords(self, duration=False):
        """ """
        res = [chord.to_chord(duration=duration) for chord in self.chords]
        return res

    def to_chords_with_duration(self):
        durations = [chord.duration for chord in self.chords]
        res = [chord.to_chord()(r.set_duration(duration)) for duration, chord in zip(durations, self.chords)]
        return res


    def copy(self):
        """
        Copy the score
        """
        return Score([c.copy() for c in self.chords], config=self.config.copy(), tags=set(self.tags))

    def o(self, val):
        """
        Apply octave to the score.

        Parameters
        ----------
        val : int
            Nb octave to transpose
            

        Returns
        -------
        score: Score
            Octaved score

        """
        return Score([c.o_melody(val) for c in self], config=self.config.copy(), tags=self.tags)

    def __add__(self, other):
        """
        Add another score, or a chord to the current score
        Returns A copy of the score

        Parameters
        ----------
        other: Store or Chord

        Returns
        -------
        score: Score

        """

        from .chord import Chord
        if other is None:
            return self.copy()
        if isinstance(other, Chord):
            return Score(self.copy().chords + [other], config=self.config.copy(), tags=self.tags)
        if isinstance(other, Score):
            return Score(self.copy().chords + other.copy().chords, config=self.config.copy(), tags=self.tags.union(other.tags))
        else:
            raise Exception('Cannot add to Score if not Chord or Score')

    def __iter__(self):
        return self.chords.__iter__()

    def __and__(self, other):
        if isinstance(other, int):
            return Score([c & other for c in self.chords], tags=self.tags)
        else:
            raise Exception(f'Not compatible type with & {other.__class__}')


    @property
    def parts(self):
        return self.instruments

    @property
    def instruments(self):
        """

        Parameters
        ----------

        Returns
        -------
        type
            :return:

        """

        def f7(seq):
            seen = set()
            seen_add = seen.add
            return [x for x in seq if not (x in seen or seen_add(x))]

        result = []
        for chord in self:
            insts = list(chord.score.keys())
            result += insts
            result = f7(result)
        return result


    def normalize(self):
        """
        Normalize the score to be standardized for the language model
        - Normalize instruments in each chord
        - Convert relative notes to real notes
        - Correct chord octaves to minimize distance with central C
        FIXME : Relative notes should be converted first

        Returns
        -------
        score: Score
        """
        import time
        start = time.time()
        score = self.to_standard_note()
        score = score.correct_chord_octave()
        score = score.normalize_instrument_names()
        score = score.split_too_long_chords(8)
        #score = score.remove_drums()
        score = score.remove_empty_chords()
        return score

    def remove_empty_chords(self):
        return Score([c for c in self.chords if c.duration > 0])

    def correct_chord_octave(self):
        """
        Correct chord octaves to minimize distance with central C
        Returns
        -------
        score: Score
        """
        return Score([c.correct_chord_octave() for c in self.chords])

    def octaver(self, **instruments_octaves):
        """
        Transpose down octave per instrument


        Parameters
        ----------
        instruments_octaves: **keywords
                key : instrument_name, val:

        Examples
        --------



        Returns
        -------
        """
        score = None
        for chord in self:
            score += chord(**{instrument: melody.o(instruments_octaves.get(instrument, 0)) for instrument, melody in chord.items()})

        return score

    def to_events(self, tempo=120, **kwargs):
        from musiclang.write.out import score_to_events
        events = score_to_events(self, tempo=tempo, **kwargs)
        return events

    def to_events_pickle(self, filepath, tempo=120, **kwargs):
        from musiclang.write.out import score_to_events
        import pickle
        events = score_to_events(self, tempo=tempo, **kwargs)
        with open(filepath, 'wb') as f:
            pickle.dump(events, f)

    def apply(self, keep_durations=True, **melodies):
        if keep_durations:
            return Score([c(**melodies).augment(c.duration) for c in self.chords])
        else:
            return Score([c(**melodies) for c in self.chords])

    def to_voicing(self, nb_voices=4, instruments=None):
        """Convert score to a four voice voicing using the extensions provided in the chord.

        It will remove the existing scores of each chord and create the associated voicings

        Parameters
        ----------
        instruments : None or List[str] (Default value = ['piano__0', 'piano__1', 'piano__2', 'piano__3'])
                      The list of instruments used to create the voicings.

        Returns
        -------

        score: Score
               The score with voicings corresponding to chords

        """
        score = None
        for chord in self:
            score += chord.to_voicing(nb_voices=nb_voices, instruments=instruments)

        return score

    def show(self, *args, **kwargs):
        """Wrapper to the music21 show method
        :return:

        Parameters
        ----------
        *args :
            
        **kwargs :
            

        Returns
        -------

        """
        import music21
        import tempfile
        import os
        with tempfile.TemporaryDirectory() as di:
            m21_score = self.to_music21(**kwargs)
            return m21_score.show(*args)


    def __setitem__(self, key, value):
        return self.chords.__setitem__(key, value)

    def __getitem__(self, item):
        """
        If str return a score with only this voice
        else returns item getter of the list of chords and convert it back to a score
        """
        from .note import Silence
        from .chord import Chord
        if isinstance(item, str):
            new_score = None
            for chord in self:
                if item in chord.score.keys():
                    new_score += chord(**{item: chord.score[item]})
                else:
                    new_score += chord(**{item: Silence(chord.duration)})
            return new_score
        elif isinstance(item, list):
            new_score = None
            for chord in self:
                chord_score = {}
                for it in item:
                    if it in chord.score.keys():
                        chord_score[it] = chord.score[it]
                    else:
                        chord_score[it] = Silence(chord.duration)
                new_score += chord(**chord_score)

            return new_score
        else:
            chords = self.chords.__getitem__(item)
            if isinstance(chords, Chord):
                return chords
            return sum(chords, None)

    def put_on_same_chord(self):
        """Take the first chord as reference,
        Put everything into the first chord preserving the note value (It will change the piece to a static harmony)

        Parameters
        ----------
        score : Score
                Input score


        Returns
        -------
        score: Score
               Score containing one chord with all the melodies

        Examples
        --------

        >>> from musiclang.library import *
        >>> score = (I % I.m)(piano__0=s0) + (V % I.m)(piano__0=s2)
        >>> score.put_on_same_chord()
        (I % I.m)(piano__0=s0 + s2)

        """
        from .time_utils import put_on_same_chord
        return put_on_same_chord(self)

    @classmethod
    def json_file_to_score(cls, filepath):
        """
        Load a score from an orchestrated json file
        Parameters
        ----------
        filepath

        Returns
        -------

        """
        import json
        with open(filepath, 'r') as f:
            data = json.load(f)

        return cls.dict_to_score(data)

    @classmethod
    def from_file(cls, filepath):
        """
        Load a score from a musiclang text file
        Parameters
        ----------
        filepath

        Returns
        -------

        """
        with open(filepath, 'r') as f:
            data = f.read()
        return cls.from_str(data)

    def to_pickle(self, filepath, create_dir=False):
        """
        Save the score into a pickle file

        Parameters
        ----------
        filepath : str
                   Filepath on which to save the score
        create_dir : bool (Default value = False)
                    Create the directory if it does not exist
        """
        import pickle
        import os
        directory = os.path.dirname(filepath)
        if create_dir and not os.path.exists(directory):
            os.makedirs(directory)
        with open(filepath, 'wb') as f:
            pickle.dump(self, f)

    @classmethod
    def generate_score(cls, n_chords=1, prompt='(', temperature=0.5, top_k=10):
        """
        Generate a musical idea from a prompt
        Parameters
        ----------
        n_chords: int
            The number of new chords to predict
        prompt: str
            The prompt to start the generation
        temperature: float
            The temperature of the prediction. The higher the more random
        top_k: int
            The number of tokens to consider for the prediction.

        Returns
        -------
        score: Score
            The generated score

        """
        from musiclang.predict.predictors import predict_score_from_huggingface
        base_chords = 0
        result = prompt
        for i in range(n_chords):
            score_str = predict_score_from_huggingface(str(result), n_chords=1, temperature=temperature,
                                                       top_k=top_k)
            result = Score.from_str(score_str)
            result = Score(result.chords[:base_chords + i + 1])
            result = result.normalize_chord_duration()

        return result

    @classmethod
    def generate_chord(cls, prompt='(', temperature=0.5, top_k=10):
        """
        Generate one chord from AI model
        Parameters
        ----------
        prompt
        temperature
        top_k

        Returns
        -------

        """
        from musiclang.predict.predictors import predict_score_from_huggingface
        score_str = predict_score_from_huggingface(prompt, n_chords=1, temperature=temperature, top_k=top_k)
        score = Score.from_str(score_str)
        return score.chords[0].normalize_chord_duration()


    def predict_score(self, n_chords=1, temperature=0.5, top_k=10):
        """
        Predict the continuation of the score given the current score

        Please note that this method is still very experimental as we are
        currently improving our musiclang language model.

        Parameters
        ----------
        n_chords: int
            The number of new chords to predict
        temperature: float
            The temperature of the prediction. The higher the more random
            To avoid syntax errors set lower than 0.75
        top_k: int
            The number of tokens to consider for the prediction.
            The higher the more random
            To avoid syntax errors set lower than 20

        Returns
        -------
        predicted_score: Score
            The predicted score

        """
        from musiclang.predict.predictors import predict_score_from_huggingface
        base_chords = len(self.chords)
        result = self
        for i in range(n_chords):
            score_str = predict_score_from_huggingface(str(result.normalize()), n_chords=1, temperature=temperature, top_k=top_k)
            result = Score.from_str(score_str)
            result = Score(result.chords[:base_chords + i + 1])
            result = result.normalize_chord_duration()

        return result


    def normalize_chord_duration(self):
        """
        Normalize the duration of each chord in the score (project on the melody with the shortest duration for each chord)
        See :func:`~Chord.normalize_chord_duration()`
        Returns
        -------
        """
        return Score([chord.normalize_chord_duration() for chord in self.chords])

    def split_too_long_chords(self, max_length):
        """
        Split the chords that are too long into smaller chords
        Parameters
        ----------
        max_length: fraction
            The maximum length of a chord

        Returns
        -------
        score: Score
            The score with shorter chords

        """

        new_score = []
        for chord in self:
            if chord.duration > max_length:
                new_score += (chord.split(max_length)).chords
            else:
                new_score.append(chord)
        return Score(new_score)

    def to_text_file(self, filepath, create_dir=False):
        """
        Save as a musiclang txt file
        Parameters
        ----------
        filepath
        create_dir: bool (Default value = False)
            Create the directory if it does not exist
        Returns
        -------

        """
        import os
        directory = os.path.dirname(filepath)
        if create_dir and not os.path.exists(directory):
            os.makedirs(directory)
        with open(filepath, 'w') as f:
            f.write(str(self))
    @classmethod
    def dict_to_score(cls, data):
        """
        Load a score from an orchestrated json file
        Parameters
        ----------
        data

        Returns
        -------

        """
        from musiclang import PartComposer
        score = PartComposer.compose(data)
        return score



    def patternize(self,
                   **kwargs):
        """
        Extract the pattern from the chord

        Parameters
        ----------
        nb_excluded_instruments: int (Default value = 0)
            Number of instruments to exclude from the pattern
        fixed_bass: bool (Default value = True)
            If True, the bass will be fixed in the pattern response
        voice_leading: bool (Default value = True)
            If True, the voice leading will be applied in the pattern response
        melody: bool (Default value = False)
            Do you want to extract a melody or an accompaniment pattern
        instruments: list[str] (Default value = None)
            List of instruments to use for the pattern
        voicing: list[Note] or None (Default value = None)
            Voicing to use for the pattern, otherwise extract the good one
        add_metadata: bool (Default value = True)
            If True, add metadata and features to the pattern
        Returns
        -------
        score_pattern: Chord
            Patternized chord or score
        pattern: dict
            Pattern data of the score

        """
        # Project all on same chord
        first_chord = self.chords[0].to_chord().augment(self.duration)
        one_chord_score = self.project_on_score(first_chord, voice_leading=False)
        chord = one_chord_score.chords[0]
        pattern, score_pattern = chord.patternize(**kwargs)
        pattern['chords'] = [str(chord.to_chord()(r.augment(chord.duration))) for chord in self.chords]
        pattern['multi_chord'] = True
        return pattern, score_pattern



    def get_patterns(self, nb_excluded_instruments=0, **kwargs):
        metadata_base = {
            "tempo": self.config['tempo']
        }
        metadata_base = {**metadata_base, **kwargs}
        data = []
        patterns = []
        for chord_idx, chord in enumerate(self.chords):
            if len(chord.parts) <= nb_excluded_instruments:
                continue
            try:
                dict_pattern, pattern = chord.patternize(nb_excluded_instruments=nb_excluded_instruments, **kwargs)
                dict_pattern['metadata'] = {**dict_pattern['metadata'], **metadata_base}
                dict_pattern['metadata']['chord_idx'] = chord_idx
                data.append(dict_pattern)
                patterns.append(pattern)
            except Exception as e:
                print("Error with chord: ", chord)
                print(e)
        return data, patterns

    def project_pattern(self, score, restart_each_chord=False):
        """

        Parameters
        ----------
        score: Score or list[Note]
            - If score : Score on which the pattern will be projected
            - If list : List of notes on which the pattern will be projected (It will replace the adequate "xi")

        Returns
        -------
        score: Score
            Score with the pattern projected on it
        """
        from musiclang.transform.composing import Patternator
        if isinstance(score, list):
            voicing = score
            parts = self.parts
            score = NC(**{part: Score.from_str(voicing[idx]).set_duration(self.duration)
                          for idx, part in enumerate(parts)})
        else:
            pass

        score_patterned = Patternator(restart_each_chord=restart_each_chord)(self, score)
        return score_patterned

    def project_on_score(self, score2, keep_pitch=False, voice_leading=True, keep_score=False,
                         repeat_to_duration=False, allow_override=False):
        """Project harmonically the current score onto the score2.
        The score2 notes will disappear, only keeping the current score with the harmony of score2
        Can either project by complete diatonic transposition or trying to move as little as possible the notes
        (voice_leading=True)

        Algorithm : For each chord of score2 : get chords that belongs to score1 and reproject on chord of score2

        Parameters
        ----------
        score : Score
                Score to project on the chords of score2
        score2 : Score
            Score that contains the harmony
        voice_leading: boolean (default=True)
            If true, we try to move the notes as little as possible
        voice_leading : boolean (default=True)
            If True try to move the notes as little as possible (Default value = True)
        repeat_to_duration: boolean (default=False)
            If True then Repeat the score to fit the duration of score2
        allow_override: boolean (default=False)
            In case keep_score is True should we allow existence of intersection of voices between self and score2 ?
            If True score2 parts that exists in self will be replaced by projected parts of self. Otherwise raise an Exception
        Returns
        -------
        new_score: Score
                  The score projected on score2 chord progression. It will fit the minimal duration of both scores

        Examples
        --------

        >>> from musiclang.library import *
        >>> score1 = (I % I.m)(piano__0=s0.e) + (V % I.m)(piano__0=s2.e)
        >>> score2 = (II % III.m)(piano__0=s0.e + s0.e)
        >>> score1.project_on_score(score2, voice_leading=False)
        (II % III.m)(piano__0=s0.e + s2.e)

        """
        # Algo : For each chord of score2 : get chords that belongs to score1 and reproject on chord of score2
        from .time_utils import project_on_score
        from musiclang.transform.composing import project_on_score_keep_notes

        if repeat_to_duration and self.duration < score2.duration:
            diff_duration = (score2.duration // self.duration) + 1
            to_project = self * diff_duration
        else:
            to_project = self

        if keep_pitch:
            to_project = self.to_absolute_note()

        if voice_leading:
            result_score = project_on_score_keep_notes(to_project.to_score(), score2.to_score())
        else:
            result_score = project_on_score(to_project.to_score(), score2.to_score(), keep_score=False)

        if keep_score:
            if not allow_override and len(set(result_score.parts).intersection(score2.parts)) != 0:
                raise Exception('If keep_score flag is True, parts should be differents between the scores')
            result_score = sum([c1(**{**c2.score, **c1.score}) for c1, c2 in zip(result_score.chords, score2.chords)], None)
        return result_score

    def project_on_rhythm(self, rhythm, **kwargs):
        return Score([chord.project_on_rhythm(rhythm, **kwargs) for chord in self.chords], tags=self.tags)

    def to_absolute_note(self):
        return Score([chord.to_absolute_note() for chord in self.chords], tags=self.tags)


    def get_chord_between(self, chord, start, end):
        """

        Parameters
        ----------
        chord :
            
        start :
            
        end :
            

        Returns
        -------

        """
        from .time_utils import get_chord_between
        return get_chord_between(chord, start, end)

    @property
    def pedal_each_chord(self):
        return Score([c.pedal for c in self.chords], tags=set(self.tags))

    def doubling(self, **args):
        new_score = []
        for chord in self.chords:
            score = chord.copy().score
            for part, new_parts in args.items():
                if isinstance(new_parts, str):
                    new_parts = [new_parts]
                for new_part in new_parts:
                    if part in score.keys():
                        score[new_part] = score[part].copy()
            new_score.append(chord(**score))

        return Score(new_score)

    def get_score_between(self, start=None, end=None):
        """
        Get the score between start and end time.

        Parameters
        ----------
        start :
             (Default value = None)
        end :
             (Default value = None)

        Returns
        -------

        """
        from .time_utils import get_score_between
        return get_score_between(self, start, end)

    def reduce(self, n_voices=4, start_low=False, instruments=None):
        """
        Reduce the score to n_voices

        Parameters
        ----------
        n_voices : int, (Default value = 4)
                   Number of voices in the reduction
        start_low : bool (Default value = False)
                    If true the first voice in the reduction will be the bass, else the soprano
        instruments : List[str] or None (Default value = None)
                      Name of the voices in the reduction
        Returns
        -------
        score : Score
                The reduced score

        Examples
        --------

        >>> from musiclang.library import *
        >>> score = (I % I.M)(piano=[s0, s2, s4]) + (V % I.M)(piano=[s0, s2, s4])
        >>> score.reduce(n_voices=2, start_low=False, instruments=['cello__0', 'violin__0'])
        (I % I.M)(
            cello__0=s0,
            violin__0=s4)+
        (V % I.M)(
            cello__0=s0,
            violin__0=s4)

        """
        from .arrange_utils import reduce
        return reduce(self, n_voices=n_voices, start_low=start_low, instruments=instruments)



    @classmethod
    def from_annotation_file(cls, file):
        with open(file, 'r') as f:
            text = f.read()
        return Score.from_annotation(text)

    @classmethod
    def from_annotation(cls, text):
        from musiclang.analyze import ScoreFormatter
        return ScoreFormatter(text).parse()


    @classmethod
    def from_midi(cls, filename, fast_chord_inference=True, chord_range=None):
        """
        Load a midi score into musiclang

        .. warning:: This step can take some time depending on the length of the file

        Parameters
        ----------
        filename : str
                   Filepath of the file

        Returns
        -------
        score: Score
               The loaded score

        """
        from musiclang.analyze import parse_to_musiclang
        score, config = parse_to_musiclang(filename, fast_chord_inference=fast_chord_inference, chord_range=chord_range)
        real_config = {**score.config, **config}
        score.config = real_config
        return score

    def to_extension_note(self):
        return Score([chord.to_extension_note() for chord in self.chords], tags=set(self.tags))

    def to_chord_note(self):
        return Score([chord.to_chord_note() for chord in self.chords], tags=set(self.tags))

    def to_standard_note(self):
        return Score([chord.to_standard_note() for chord in self.chords], tags=set(self.tags))


    @classmethod
    def from_xml(cls, filename, fast_chord_inference=False):
        """
        Load a musicxml score into musiclang

        .. warning:: This step can take some time depending on the length of the file


        Parameters
        ----------
        filename : str
                   Filepath of the file
            

        Returns
        -------

        score: Score
               The loaded score

        """
        from musiclang.analyze import parse_to_musiclang
        score, config = parse_to_musiclang(filename, fast_chord_inference=fast_chord_inference)
        score.config.update(config)
        return score

    def decompose_duration(self):
        """
        Decompose the duration in a note + continuations. It recursively call
        :func:`~Note.decompose_duration()`
        """
        return Score([chord.decompose_duration() for chord in self.chords], tags=self.tags)

    def to_score(self, copy=True):
        """ """
        if not copy:
            return self
        else:
            return self.copy()


    @classmethod
    def from_str(cls, s):
        try:
            from musiclang import Chord
            pattern = re.compile(r'(?<=\))\s*\+\s*(?=\()')
            data = re.split(pattern, str(s))
            chords = [eval(str(d).replace('\n', '')) for d in data]
            assert isinstance(chords[0], Chord)
            result = Score(chords)
            if len(result.chords) == 1:
                return result.chords[0]
            return result
        except:
            return eval(str(s).replace('\n', ''))

    @classmethod
    def from_sequence(cls, sequence, **kwargs):
        """

        Parameters
        ----------
        sequence :
            
        **kwargs :
            

        Returns
        -------

        """
        from .sequence.sequence import sequence_to_score
        return sequence_to_score(sequence, **kwargs)

    def to_sequence(self, **kwargs):
        """
        Convert
        Parameters
        ----------
        **kwargs :
            

        Returns
        -------

        """
        from .sequence.sequence import score_to_sequence
        return score_to_sequence(self, **kwargs)


    @classmethod
    def from_pickle(cls, filepath):
        """

        Parameters
        ----------
        filepath :
            

        Returns
        -------

        """
        import pickle
        with open(filepath, 'rb') as f:
            data = pickle.load(f)
        return data

    def __eq__(self, other):
        from .chord import Chord
        if isinstance(other, Chord):
            return self == Score([other], config=self.config.copy())
        if not isinstance(other, Score):
            return False
        elif len(other.chords) != len(self.chords):
            return False
        else:
            return all([c1 == c2 for c1, c2 in zip(self.chords, other.chords)])

    def __getattr__(self, item):
        chords = self.copy()
        chords.chords = [getattr(s, item) for s in self.chords]
        return chords

    def __mod__(self, other):
        from .tonality import Tonality
        if isinstance(other, Tonality):
            return Score([c % other for c in self.chords], config=self.config.copy(), tags=self.tags)
        else:
            raise Exception('Following % should be a Tonality')

    def remove_accidents(self):
        return Score([chord.remove_accidents() for chord in self.chords], tags=set(self.tags))

    def remove_drums(self):
        return Score([chord.remove_drums() for chord in self.chords])


    def __mul__(self, other):
        """
        If other is Integer, repeat the note other times
        """
        if isinstance(other, int):
            return sum([self.copy() for i in range(other)], None)
        else:
            raise Exception('Cannot multiply Score and ' + str(type(other)))

    def __radd__(self, other):
        if other is None:
            return self.copy()

    def __repr__(self):
        return '+ \n'.join([str(chord) for chord in self.chords])

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__ = d


    @property
    def duration(self):
        """ """
        return sum([c.duration for c in self.chords])

    def to_code(self, **kwargs):
        """Export the chord serie as a string representing valid python code that recreates the score
        :return:

        Parameters
        ----------
        **kwargs :
            

        Returns
        -------

        """
        from .out.to_code import chord_serie_to_code

        code = chord_serie_to_code(self, **kwargs)
        return code


    def realize_tags(self):
        new_score = None
        last_notes = {}
        final_notes = {}
        for idx, chord in enumerate(self.chords):
            if idx > 0:
                last_notes = {instrument: self.chords[idx-1].score[instrument].notes[-1] for instrument in self.chords[idx-1].score.keys()}
            if idx < len(self.chords)-1:
                final_notes = {instrument: self.chords[idx+1].score[instrument].notes[0] for instrument in self.chords[idx+1].score.keys()}

            new_score += chord.realize_tags(last_note=last_notes, final_note=final_notes)

        return new_score

    def to_code_file(self, filepath, **kwargs):
        """Export the chord serie as a file representing valid python code that recreates the score

        Parameters
        ----------
        filepath :
            return:
        **kwargs :
            

        Returns
        -------

        """
        code = self.to_code(**kwargs)
        with open(filepath, 'w') as f:
            f.write(code)

    def to_musicxml(self, filepath, signature=(4, 4), tonality=None, tempo=120, no_repeat=False, **kwargs):
        """
        Transform a musiclang score into a musicxml file, readable by all the main notation software (musescore, finale ...)

        Parameters
        ----------
        score: Score
                MusicLang score to transform
        filepath: str
                Filepath of the musicxml file
        signature: tuple (nom, den)
                Time signature of the piece
        tempo: int
                Tempo of the piece
        tonality: Tonality
                Tonality of the piece if applicable
        title: str
            Title of the piece

        """
        # Convert score to midi
        from .out.to_mxl import score_to_mxl

        return score_to_mxl(self, filepath, signature=signature, tonality=tonality, tempo=tempo, no_repeat=no_repeat, **kwargs)

    def to_music21(self, signature=(4, 4), tonality=None, tempo=120, **kwargs):
        """
        Transform a musiclang score into a Music21 score
        Parameters
        ----------
        score: Score
                MusicLang score to transform
        signature: tuple (nom, den)
                Time signature of the piece
        tempo: int
                Tempo of the piece
        tonality: Tonality
                Tonality of the piece if applicable
        title: str
            Title of the piece

        Returns
        -------
        m21_score: music21.Score
            Score transformed into a music21 object

        """
        from .out.to_mxl import score_to_music_21
        return score_to_music_21(self, signature=signature, tonality=tonality, tempo=tempo, **kwargs)

    def to_midi(self, filepath, **kwargs):
        """
        Save the score to midi

        Parameters
        ----------
        filepath :
            
        **kwargs :
            

        Returns
        -------

        """
        # Convert score to midi
        from .out.to_midi import score_to_midi

        return score_to_midi(self, filepath, **kwargs)


    def to_scale_notes(self):

        return Score([chord.to_scale_notes() for chord in self.chords])


    def to_custom_chords(self, nb_voices=4, fixed_bass=True, **kwargs):

        from musiclang.transform import VoiceLeading
        from musiclang import CustomChord
        chords = Score([c(*[Note("b", i, 0, c.duration) for i in range(nb_voices)]) for c in self.chords])
        fixed_voices = ['piano__0'] if fixed_bass else None
        score = VoiceLeading(fixed_voices=fixed_voices)(Score(chords), **kwargs)
        score = score.to_scale_notes()
        return Score([CustomChord(list(sorted([chord.set_degree(0).set_octave(0).parse(chord.to_pitch(n.notes[0]))
                                               for n in chord.score.values()], key=lambda x: chord.to_pitch(x))),
                                  tonality=chord.tonality).set_duration(chord.duration)
                      for chord in score.chords])

    @classmethod
    def from_orchestration(cls, orchestration, chords):
        def parse_one_data(ins, data, score, chord, rhythm=None, idx=None):
            notes = parse_notes(chord, data['note'], data['octave'])
            instrument_name = ins if idx is None else f'{ins}__{idx}'
            score[instrument_name] =  parse_pattern(data['pattern']).apply_pattern(*notes)
            if chord.duration != 0:
                score[instrument_name] = score[instrument_name].get_between(0, chord.duration)
            if rhythm is not None:
                score[instrument_name] = apply_rhythm(score[instrument_name], rhythm)
            return score

        def parse_pattern(pattern):

            if isinstance(pattern, str):
                return Score.from_str(pattern)
            else:
                return pattern
        def parse_notes(chord, note, octave):
            if isinstance(note, str):
                return chord.__getattribute__(note).o(octave)
            else:
                return [chord.__getattribute__(n).o(octave) for n in note]

        def parse_dict(orchestration, chord):
            score = {}

            for ins, data in orchestration.items():
                if isinstance(data, dict):
                    score = parse_one_data(ins, data, score, chord, idx=None)
                else:
                    for idx, d in enumerate(data):
                        score = parse_one_data(ins, d, score, chord, idx=idx)
            return score

        def apply_rhythm(melody, rhythm):
            # Create rhythm
            from musiclang import Metric
            metric = Metric(rhythm['rhythm'], signature=rhythm['time_signature'], tatum=rhythm['tatum'])
            return metric.apply_to_melody(melody)


        def parse_list(orchestration, chord):
            score = {}
            instrument_counter = {}
            for data in orchestration:
                if isinstance(data, dict):
                    ins = data['instrument']
                    instrument_counter[ins] = instrument_counter.get(ins, -1) + 1
                    idx = instrument_counter[ins]
                    score = parse_one_data(ins, data, score, chord, rhythm=data.get('rhythm', None), idx=idx)

                else:
                    for i, d in enumerate(data):
                        ins = d['instrument']
                        instrument_counter[ins] = instrument_counter.get(ins, -1) + 1
                        idx = instrument_counter[ins]
                        score = parse_one_data(ins, d, score, chord, rhythm=data.get('rhythm', None), idx=idx)
            return score

        all_chords = []
        for chord in chords:
            if isinstance(orchestration, dict):
                score = parse_dict(orchestration, chord)
            else:
                score = parse_list(orchestration, chord)
            all_chords.append(chord(**score))
        final_score = Score(all_chords)
        return final_score



from .element import Formatter

class BeatElement(Formatter):

    def __init__(self, time):
        self.time = time

    def to_text(self):
        return ' ' + str(self)

    def __repr__(self):
        return f'b{self.time}'


from musiclang.library import *
from musiclang import Score
import pytest


def test_drum_note_dont_octave():
    assert sn.o(1) == sn


def test_drum_note_dont_transpose():
    assert (sn & 2) == sn


def test_drums_auto_converted():

    chord = (I % I.M)(drums=s0.o(-2))
    assert chord == (I % I.M)(drums=bd)

def test_unexisting_drum_lead_to_error():

    chord = (I % I.M)(drums=s0.o(-2) + s0)
    assert chord == (I % I.M)(drums=s0.o(-2) + d0)

def test_drum_note_evaluation_ok():
    score = (I % I.M)(drums=[sn + bd, hh + hh]) + (I % I.M)(drums=[sn + bd, hh + hh])
    assert Score.from_str(str(score)) == score


"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""
from functools import cached_property

from .constants import *
from .chord import Chord
import re


class CustomChord(Chord):
    """
    Represents a custom chord in MusicLang.

    The custom chord is composed of :
        - One set of notes
        - One tonality itself composed of :
            - A degree in ``I, II, III, IV, V, VI, VII``
            - A mode in ``'M', 'm', 'mm', 'dorian, 'phrygian', 'lydian', 'mixolydian', 'aeolian', 'locrian'``
        - One score : A Dictionary where keys are part names, values are :func:`~musiclang.Melody`
        - One octave : An integer that specifies on which octave transpose the chord

    .. note:: In reality a chord in MusicLang is a "chord scale" because it always implies a tonality.

    Examples
    --------

    >>> from musiclang.library import *
    >>> custom_chord = I.M(s0, s2, s4)(piano__0=b0.h + b1.h)
    >>> custom_chord
    I.M(s0, s2, s4)(piano__0=b0.h + b1.h)
    """
    def __init__(self, notes, extension='', tonality=None, score=None, octave=0, tags=None, **kwargs):
        """

        Parameters
        ----------
        notes: List[int]
            Notes that compose the chord
        extension: int or str , optional
                 Represents the figured bass. Must be in ['','5','6','64','7','65','43','2','9','11','13']
        tonality: musiclang.Tonality, optional
                Represents the tonality of the chord
        score: dict, optional
               Represents the melodies and instruments played by this chord
        octave: int, optional
                Represents the base octave of the chord
        """
        super().__init__(0, extension=extension, tonality=tonality, score=score, octave=octave, tags=tags)
        self.notes = notes


    @property
    def possible_notes(self):
        """
        List the notes that belong to the chord (only the chord, not the scale)

        Returns
        -------
        notes: List[Note]
               The list of possible notes for a chord

        """
        return [n.copy() for n in self.notes]


    def tonality_to_str(self):
        """
        Convert the tonality of the chord to a string.

        Parameters
        ----------

        Returns
        -------

        """
        if self.tonality is None:
            return "I.M"

        return self.tonality.to_code()



    def copy(self):
        """
        Copy the current custom chord
        """
        return CustomChord(
                     notes=self.notes,
                     extension=self.extension,
                     tonality=self.tonality.copy() if self.tonality is not None else None,
                     score={k: s.copy() for k, s in self.score.items() if
                            s is not None} if self.score is not None else None,
                     octave=self.octave,
                     tags=set(self.tags)
                     )

    def __repr__(self):
        return f"{self.to_code()}({self.melody_to_str()})"


    @cached_property
    def chord_notes(self):
        return [c.copy() for c in self.notes]

    @cached_property
    def extension_notes(self):
        # Try a composite with a match
        return [c.copy() for c in self.notes]

    def notes_to_str(self):
        return  ",".join([str(c) for c in self.notes])

    def to_code(self):
        """Export the code as python interpretable str.
        It ignores the score
        :return:

        Parameters
        ----------

        Returns
        -------

        """
        chord_str = f"{self.tonality_to_str()}({self.notes_to_str()})"
        if self.octave != 0:
            chord_str = f"{chord_str}.o({self.octave})"

        return chord_str




"""Several cached music21 objects and functions for performance."""

from music21.key import Key
from music21.pitch import Pitch
from music21.interval import Interval
from .keydistance import weberEuclidean as _we
from .keydistance import getTonicizationScaleDegree as _gtsd

_transposeKey = {}
_transposePitch = {}
_transposePcSet = {}
_pitchObj = {}
_keyObj = {}
_intervalObj = {}
_weberEuclidean = {}
_getTonicizationScaleDegree = {}


def weberEuclidean(k1, k2):
    """A cached version of keydistance.weberEuclidean.

    Parameters
    ----------
    k1 :
        
    k2 :
        

    Returns
    -------

    """
    duple = (k1, k2)
    if duple in _weberEuclidean:
        return _weberEuclidean[duple]
    distance = _we(k1, k2)
    _weberEuclidean[duple] = distance
    return distance


def getTonicizationScaleDegree(localKey, tonicizedKey):
    """A cached version of keydistance.weberEuclidean.

    Parameters
    ----------
    localKey :
        
    tonicizedKey :
        

    Returns
    -------

    """
    duple = (localKey, tonicizedKey)
    if duple in _getTonicizationScaleDegree:
        return _getTonicizationScaleDegree[duple]
    degree = _gtsd(localKey, tonicizedKey)
    _getTonicizationScaleDegree[duple] = degree
    return degree


def forceTonicization(localKey, candidateKeys):
    """Forces a tonicization of candidateKey that exist in vocabulary.

    Parameters
    ----------
    localKey :
        
    candidateKeys :
        

    Returns
    -------

    """
    tonicizationDistance = 1337
    tonicization = ""
    for candidateKey in candidateKeys:
        distance = weberEuclidean(localKey, candidateKey)
        # print(f"\t{localKey} -> {candidateKey} = {distance}")
        scaleDegree = getTonicizationScaleDegree(localKey, candidateKey)
        if scaleDegree not in ["i", "III"]:
            # Slight preference for parallel minor and relative major
            distance *= 1.05
        if scaleDegree not in ["i", "I", "III", "iv", "IV", "v", "V"]:
            distance *= 1.05
        if distance < tonicizationDistance:
            tonicization = candidateKey
            tonicizationDistance = distance
    return tonicization


def TransposeKey(key, interval):
    """Transposes a key based on an interval string (e.g., 'm3').

    Parameters
    ----------
    key :
        
    interval :
        

    Returns
    -------

    """
    duple = (key, interval)
    if duple in _transposeKey:
        return _transposeKey[duple]
    keyObj = m21Key(key)
    transposed = keyObj.transpose(interval).tonicPitchNameWithCase
    _transposeKey[duple] = transposed
    return transposed


def TransposePitch(pitch, interval):
    """Transposes a pitch based on an interval string (e.g., 'm3').

    Parameters
    ----------
    pitch :
        
    interval :
        

    Returns
    -------

    """
    duple = (pitch, interval)
    if duple in _transposePitch:
        return _transposePitch[duple]
    pitchObj = m21Pitch(pitch)
    transposed = pitchObj.transpose(interval).nameWithOctave
    _transposePitch[duple] = transposed
    return transposed


def TransposePcSet(pcset, interval):
    """Transposes a pcset based on an interval string (e.g., 'm3').

    Parameters
    ----------
    pcset :
        
    interval :
        

    Returns
    -------

    """
    duple = (pcset, interval)
    if duple in _transposePcSet:
        return _transposePcSet[duple]
    semitones = m21IntervalStr(interval).semitones
    transposed = [(x + semitones) % 12 for x in pcset]
    transposed = tuple(sorted(transposed))
    _transposePcSet[duple] = transposed
    return transposed


def m21IntervalStr(interval):
    """A cached interval object, based on the string (e.g., 'm3').

    Parameters
    ----------
    interval :
        

    Returns
    -------

    """
    if interval in _intervalObj:
        return _intervalObj[interval]
    intervalObj = Interval(interval)
    _intervalObj[interval] = intervalObj
    return intervalObj


def m21Interval(pitch1, pitch2):
    """A cached interval object, computed from two pitches.

    Parameters
    ----------
    pitch1 :
        
    pitch2 :
        

    Returns
    -------

    """
    duple = (pitch1, pitch2)
    if duple in _intervalObj:
        return _intervalObj[duple]
    p1, p2 = m21Pitch(pitch1), m21Pitch(pitch2)
    intervalObj = Interval(p1, p2)
    _intervalObj[duple] = intervalObj
    return intervalObj


def m21Key(key):
    """A cached key object, based on a string (e.g., 'c#').

    Parameters
    ----------
    key :
        

    Returns
    -------

    """
    if key in _keyObj:
        return _keyObj[key]
    keyObj = Key(key)
    _keyObj[key] = keyObj
    return keyObj


def m21Pitch(pitch):
    """A cached pitch object, based on a string (e.g., 'C#').

    Parameters
    ----------
    pitch :
        

    Returns
    -------

    """
    if pitch in _pitchObj:
        return _pitchObj[pitch]
    pitchObj = Pitch(pitch)
    _pitchObj[pitch] = pitchObj
    return pitchObj



from musiclang.library import *
from musiclang.transform.composing import VoiceLeading



def test_voice_leading_use_fixed_voices():
    score = (I % I.M)(cello=s0, violin=s1) * 2
    vl = VoiceLeading(
        fixed_voices=['cello__0'],
        seed=5,
        method='voices_and_rules',
        max_iter=10,
        max_iter_rules=10,
        temperature=4,
        min_temp=1,
        max_norm=3
    )
    new_score = vl.optimize(
        score,
        )


    assert new_score == (I % I.M)(cello=s0, violin=s3) + (I % I.M)(cello=s0, violin=s1)


def test_voice_leading_use_types():
    score = (I % I.M)(cello=s0, violin=h2) + (I % I.M)(cello=s0, violin=h2)
    vl = VoiceLeading(
        types=['s'],
        seed=1,
        method='rules',
        max_iter=100,
        max_iter_rules=10,
        temperature=4,
        min_temp=1,
        max_norm=1
    )
    new_score = vl.optimize(score)

    assert new_score == (I % I.M)(cello__0=s5.o(-1), violin__0=h2)+ (I % I.M)(cello__0=s0,	violin__0=h2)

def test_voice_leading_correct_parallel_dissonnances():
    score = (I % I.M)(cello=s0, violin=s1) + (I % I.M)(cello=s0, violin=s1)
    vl = VoiceLeading(
        seed=3,
        fixed_voices=['violin__0'],
        method='rules',
        max_iter=100,
        max_iter_rules=10,
        temperature=4,
        min_temp=1,
        max_norm=3
    )
    new_score = vl.optimize(score)

    assert new_score == (I % I.M)(	cello__0=s5.o(-1),	violin__0=s1)+ (I % I.M)(cello__0=s0,	violin__0=s1)

def test_voice_leading_silences_works():
    score = (I % I.M)(cello=s0, violin=r) + (I % I.M)(cello=s0, violin=r)
    vl = VoiceLeading(
        seed=3,
        fixed_voices=['violin__0'],
        method='rules',
        max_iter=100,
        max_iter_rules=10,
        temperature=4,
        min_temp=1,
        max_norm=1
    )
    new_score = vl.optimize(score)

    assert new_score == (I % I.M)(cello__0=s4.o(-1), violin__0=r)+ (I % I.M)(cello__0=s5.o(-1),	violin__0=r)


"""An auxiliary module to measure key distance."""

import music21
import numpy as np

# The diagonal of keys in the Weber tonal chart
# constrained to the range of keys between [Fb, G#]
WEBERDIAGONAL = [
    "B--",
    "c-",
    "F-",
    "g-",
    "C-",
    "d-",
    "G-",
    "a-",
    "D-",
    "e-",
    "A-",
    "b-",
    "E-",
    "f",
    "B-",
    "c",
    "F",
    "g",
    "C",
    "d",
    "G",
    "a",
    "D",
    "e",
    "A",
    "b",
    "E",
    "f#",
    "B",
    "c#",
    "F#",
    "g#",
    "C#",
    "d#",
    "G#",
    "a#",
    "D#",
    "e#",
    "A#",
    "b#",
]

# Adding this vector takes you to the next coordinates of a key
# in the Weber tonal chart, ascending from "flatness" to "sharpness"
TRANSPOSITION = np.array((2, 3))


def weberEuclidean(k1, k2):
    """A measurement of key distance based on the Weber tonal chart.

    Parameters
    ----------
    k1 :
        
    k2 :
        

    Returns
    -------

    """
    i1, i2 = WEBERDIAGONAL.index(k1), WEBERDIAGONAL.index(k2)
    flatterkey, sharperkey = sorted((i1, i2))
    coord1 = np.array((flatterkey, flatterkey))
    coord2 = np.array((sharperkey, sharperkey))
    smallerdistance = 1337
    for i in range(len(WEBERDIAGONAL) // 2):
        trans = TRANSPOSITION * i
        newcoord1 = trans + coord1
        distance = np.linalg.norm(coord2 - newcoord1)
        if distance < smallerdistance:
            smallerdistance = distance
    return smallerdistance


def getTonicizationScaleDegree(localKey, tonicizedKey):
    """

    Parameters
    ----------
    localKey :
        
    tonicizedKey :
        

    Returns
    -------
    type
        

    """
    tonic, _, third, _, fifth, _, _, _ = music21.key.Key(tonicizedKey).pitches
    c1 = music21.chord.Chord([tonic, third, fifth])
    # TODO: Use harmalysis to solve this problem, not romanNumeralFromChord
    degree = music21.roman.romanNumeralFromChord(c1, localKey).figure
    # TODO: This is a hack to workaround music21
    if localKey.islower() and degree == "bVI":
        degree = "VI"
    return degree



"""Various utilities used throughout the other modules."""

import os

import numpy as np


def tensorflowGPUHack():
    """ """
    import tensorflow as tf

    # https://github.com/tensorflow/tensorflow/issues/37942
    gpu_devices = tf.config.experimental.list_physical_devices("GPU")
    for device in gpu_devices:
        tf.config.experimental.set_memory_growth(device, True)


def disableGPU():
    """ """
    # Disabling the GPU
    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"


def padToSequenceLength(arr, sequenceLength, value=0):
    """

    Parameters
    ----------
    arr :
        
    sequenceLength :
        
    value :
         (Default value = 0)

    Returns
    -------

    """
    frames, features = arr.shape
    featuresPerSequence = sequenceLength * features
    featuresInExample = frames * features
    padding = featuresPerSequence - (featuresInExample % featuresPerSequence)
    paddingTimesteps = int(padding / features)
    arr = np.pad(arr, ((0, paddingTimesteps), (0, 0)), constant_values=value)
    arr = arr.reshape(-1, sequenceLength, features)
    return arr


class DynamicArray:
    """ """
    def __init__(
        self,
        shape=(0,),
        dtype=float,
        initial_capacity=1000,
        growth_factor=2,
        memmap="",
    ):
        """First item of shape is ignored, the rest defines the shape."""
        self.shape = shape
        if memmap:
            self.data = np.memmap(
                memmap,
                mode="w+",
                shape=(initial_capacity, *shape[1:]),
                dtype=dtype,
            )
        else:
            self.data = np.zeros((initial_capacity, *shape[1:]), dtype=dtype)
        self.capacity = initial_capacity
        self.growth_factor = growth_factor
        self.memmap = memmap
        self.size = 0
        self.dtype = dtype

    def update(self, x):
        """

        Parameters
        ----------
        x :
            

        Returns
        -------

        """
        if self.size == self.capacity:
            self.capacity = int(self.capacity * self.growth_factor)
            if self.memmap:
                self.data.flush()
                self.data = np.memmap(
                    self.memmap,
                    mode="r+",
                    shape=(self.capacity, *self.data.shape[1:]),
                    dtype=self.dtype,
                )
            else:
                self.data = np.resize(
                    self.data, (self.capacity, *self.data.shape[1:])
                )
        self.data[self.size] = x
        self.size += 1

    def finalize(self):
        """ """
        return self.data[: self.size]



"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

from .constants import *

class Melody:
    """
    Main class to write a melody in MusicLang. A melody is a serie of notes

    Examples
    --------

    For example here is an example of a melody that plays the seven note of a scale :

    >>> from musiclang.write.library import *
    >>> melody =  s0 + s1 + s2 + s3 + s4 + s5 + s6
    >>> melody
    s0 + s1 + s2 + s3 + s4 + s5 + s6

    You can also create it using the melody class :

    >>> from musiclang import Melody
    >>> melody = Melody([s0, s1, s2, s3, s4, s5, s6])
    >>> melody
    s0 + s1 + s2 + s3 + s4 + s5 + s6

    """
    def __init__(self, notes, nb_bars=1, tags=None):
        from .note import Note
        if isinstance(notes, Note):
            notes = []
        self.notes = notes
        self.tags = set(tags) if tags is not None else set()
        self.nb_bars = nb_bars

    def has_tag(self, tag):
        """
        Check if the tag exists for this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        melody: Melody
        """
        return tag in self.tags


    def add_tag(self, tag):
        """
        Add a tag to this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        melody: Melody
        """
        cp = self.copy()
        cp.tags.add(tag)
        return cp

    def add_tags(self, tags):
        """
        Add several tags to the object.
        Returns a copy of the object

        Parameters
        ----------
        tags: List[str]
        tags to add

        Returns
        -------
        melody: Melody

        """
        cp = self.copy()
        cp.tags = cp.tags.union(set(tags))
        return cp

    def remove_tags(self, tags):
        """
        Remove several tags from the object.
        Returns a copy of the object

        Parameters
        ----------
        tags: List[str]

        Returns
        -------
        melody: Melody


        """
        cp = self.copy()
        cp.tags = cp.tags - set(tags)
        return cp

    def remove_tag(self, tag):
        """
        Remove a tag from this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        melody: Melody
        """
        cp = self.copy()
        cp.tags.remove(tag)
        return cp

    def clear_tags(self):
        """
        Clear all tags from this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        melody: Melody
        """
        cp = self.copy()
        cp.tags = set()

        return cp

    def set_amp(self, amp):
        return Melody([n.set_amp(amp) for n in self.notes], nb_bars=self.nb_bars, tags=set(self.tags))


    def set_duration(self, duration):
        return self.augment(duration / self.duration)

    def to_drum(self):
        return Melody([n.to_drum() for n in self.notes], nb_bars=self.nb_bars, tags=set(self.tags))

    def to_melody(self):
        return self.copy()

    def remove_accidents(self):
        return Melody([n.remove_accidents() for n in self.notes], nb_bars=self.nb_bars, tags=set(self.tags))

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)

    def __add__(self, other):
        from .note import Note
        if other is None:
            return self.copy()
        if isinstance(other, Note):
            return Melody(self.notes + [other], nb_bars=self.nb_bars, tags=set(self.tags))
        if isinstance(other, Melody):
            return Melody(self.notes + other.notes, nb_bars=self.nb_bars, tags=set(self.tags).union(other.tags))

    def __or__(self, other):
        from .note import Note
        if other is None:
            return self.copy()
        if isinstance(other, Note):
            assert other.duration == self.duration / self.nb_bars, f"Invalid duration of bar {other}: {other.duration}q, expected : {self.duration / self.nb_bars}q/bar"
            return Melody(self.notes + [other], nb_bars=self.nb_bars + 1, tags=set(self.tags))
        if isinstance(other, Melody):
            assert other.duration / other.nb_bars == self.duration / self.nb_bars, \
                f"Invalid duration of bar {other} : {other.duration}q with {other.nb_bars} bars, expected : {self.duration / self.nb_bars}q/bar"
            return Melody(self.notes + other.notes, nb_bars=self.nb_bars + other.nb_bars, tags=set(self.tags).union(other.tags))
        else:
            raise Exception(f'Invalid type when adding melody, {other.__class__}')

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(self, other):
        from .note import Note
        if isinstance(other, Note):
            return self.__eq__(Melody([other]))
        return isinstance(other, Melody) and str(other) == str(self)

    def to_absolute_note(self, chord):
        return Melody([n.to_absolute_note(chord) for n in self.notes], nb_bars=self.nb_bars, tags=set(self.tags))


    def get_between(self, start, end):
        """
        Get the notes between start and end time
        Parameters
        ----------
        start: int
        end: int

        Returns
        -------
        melody: Melody
        """
        from musiclang.write.time_utils import get_melody_between
        return get_melody_between(self, start, end)
    def get_pitches(self, chord, track_idx, time, last_note_array=None):
        """

        Parameters
        ----------
        chord :
            
        track_idx :
            
        time :
            
        last_note_array :
             (Default value = None)

        Returns
        -------

        """
        pitches = []
        for note in self.notes:
            result = note.pitch(chord, track_idx, time, last_note_array)
            if not last_note_array[SILENCE] and not last_note_array[CONTINUATION]:
                last_note_array = result

            pitches.append(result)
        return pitches

    def apply_pattern(self, *voicing):
        """
        Replace pattern notes by voicing

        Examples
        ---------

        >>> melody = x0.h + x1.h
        >>> melody.apply_pattern(s1, s0)
        s1.h + s0.h

        Parameters
        ----------
        voicing: List[Note] : Voicing used to replace in the template

        Returns
        -------
        melody: Patternized melody
        """
        new_melody =[]
        for note in self.notes:
            if note.type == "x":
                try:
                    new_melody.append(voicing[note.val].set_duration(note.duration).o(note.octave).set_amp(note.amp))
                except:
                    raise ValueError(f'Unexisting voicing index : {note.val}')
            else:
                new_melody.append(note)

        return Melody(new_melody)


    def to_scale_notes(self, chord):
        return Melody([n.to_scale_note(chord) for n in self.notes], nb_bars=self.nb_bars, tags=set(self.tags))

    def decompose_duration(self):
        """ """
        return Melody(sum([note.decompose_duration() for note in self.notes], None).notes, nb_bars=self.nb_bars, tags=set(self.tags))

    def replace_pitch(self, to_replace, new_note):
        """

        Parameters
        ----------
        to_replace :
            
        new_note :
            

        Returns
        -------

        """

        new_melody = []
        for note in self.notes:
            to_add = note.copy()
            if note.val == to_replace.val and note.type == to_replace.type:
                to_add.type = new_note.type
                to_add.val = new_note.val
            new_melody.append(to_add)
        return sum(new_melody, None)

    def replace(self, to_replace, new_note, **kwargs):
        """

        Parameters
        ----------
        to_replace :

        new_note :


        Returns
        -------

        """

        return Melody([n.replace(to_replace, new_note, **kwargs) for n in self.notes],
                      nb_bars=self.nb_bars, tags=set(self.tags))

    def set_tempo(self, tempo):
        new_melody = self.copy()
        new_melody.notes[0] = new_melody.notes[0].set_tempo(tempo)
        return new_melody


    def project_on_rhythm(self, rhythm, chord=None, **kwargs):
        from musiclang.transform.composing import project_on_rhythm
        return project_on_rhythm(rhythm, self, chord=chord)

    def accelerando(self, start, end):
        """
        Accelerate or decelerate melody from start tempo to end tempo
        Parameters
        ----------
        start: int
            Starting tempo
        end: int
            Ending tempo

        """
        new_melody = self.copy()
        L = len(new_melody)
        for i in range(len(new_melody)):
            local_tempo = int(end * (i / L) + start * ((L - i)/L))
            new_melody.notes[i] = new_melody.notes[i].set_tempo(local_tempo)
        return new_melody

    def realize_tags(self, last_note=None, final_note=None):
        new_melody = None
        for idx, note in enumerate(self.notes):
            last_note = self.notes[idx - 1] if idx - 1 >= 0 else last_note
            next_note = self.notes[idx + 1] if idx + 1 < len(self.notes) else None
            if idx == len(self.notes) - 1:
                next_note = final_note
            new_melody += note.realize_tags(last_note=last_note, next_note=next_note)

        new_melody.nb_bars = self.nb_bars
        new_melody.tags = self.tags
        return new_melody

    @property
    def pedal_on(self):
        new_melody = self.copy()
        new_melody.notes[0] = new_melody.notes[0].pedal_on
        return new_melody

    @property
    def pedal(self):
        """
        Apply pedal on first note and release on last
        """
        new_melody = self.copy()
        new_melody.notes[0] = new_melody.notes[0].pedal_on
        new_melody.notes[-1] = new_melody.notes[-1].pedal_off
        return new_melody

    @property
    def pedal_off(self):
        new_melody = self.copy()
        new_melody.notes[0] = new_melody.notes[0].pedal_off
        return new_melody


    def remove_effects(self):
        """
        Remove pedals and tempo change
        Returns
        -------
        """
        return Melody([n.remove_effects() for n in self.notes], nb_bars=self.nb_bars, tags=set(self.tags))

    def remove_tempo(self):
        """
        Remove pedals and tempo change
        Returns
        -------
        """
        return Melody([n.remove_tempo() for n in self.notes], nb_bars=self.nb_bars, tags=set(self.tags))

    def remove_pedal(self):
        """
        Remove pedals and tempo change
        Returns
        -------
        """
        return Melody([n.remove_pedal() for n in self.notes], nb_bars=self.nb_bars, tags=set(self.tags))


    def to_sequence(self, chord, inst):
        """Transform in a list of [(start_time, end_time, pitch, self)]
        :return:

        Parameters
        ----------
        chord :
            
        inst :
            

        Returns
        -------

        """
        time = 0
        sequence = []
        for note in self.notes:
            pitch = chord.to_pitch(note)
            start = time
            end = time + note.duration
            sequence.append([start, end, pitch, chord.to_chord(), inst, note])
            time += note.duration
        return sequence


    def to_standard_note(self, chord):
        return Melody([n.to_standard_note(chord) for n in self.notes], nb_bars=self.nb_bars, tags=set(self.tags))

    def clear_note_tags(self):
        return sum([n.clear_note_tags() for n in self.notes], None)

    def to_extension_note(self, chord):
        return Melody([n.to_extension_note(chord) for n in self.notes], nb_bars=self.nb_bars, tags=set(self.tags))

    def to_chord_note(self, chord):
        return Melody([n.to_chord_note(chord) for n in self.notes], nb_bars=self.nb_bars, tags=set(self.tags))


    def get_note_times(self):
        t = 0
        times = []
        for n in self.notes:
            if n.is_note or n.type in ["x", "d"]:
                times.append(t)
            t += n.duration
        return times


    def get_playing_style(self):
        """
        Get if legato, marcato or staccato by  only looking at the note durations
        Returns
        -------

        """

    def quantize_melody(self, max_frac=4):
        import math

        def lcm(a, b):
            return (a * b) // math.gcd(a, b)

        def duration_to_ts(duration):
            if float(duration) % 1 == 0:
                return int(duration), 4
            elif float(duration) % 0.5 == 0:
                return int(duration * 2), 8
            elif float(duration) % 0.25 == 0:
                return int(duration * 4), 16
            else:
                return int(duration * 8), 32
        times = self.get_note_times() # Fractions
        # Find common denominator between times
        common_denominator = 1
        for t in times:
            common_denominator = lcm(common_denominator, t.denominator)

        common_denominator = min(max_frac, common_denominator)
        step = frac(1, common_denominator)

        nb_steps = self.duration / step

        # Projects each note time (times) on the grid
        new_times = [round(t / step) * step for t in times]

        # Get index on the grid
        new_times = [int(t / step) for t in new_times]

        # As binary vector
        rhythm = [1 if i in new_times else 0 for i in range(int(nb_steps))]

        return {'rhythm': rhythm, 'tatum': (step.numerator, step.denominator),
                'time_signature': duration_to_ts(self.duration)}


    def to_code(self):
        """ """
        return " + ".join([n.to_code() for n in self.notes])

    def convert_to_drum_note(self, chord):
        return Melody([n.convert_to_drum_note(chord) for n in self.notes], nb_bars=self.nb_bars, tags=set(self.tags))

    @property
    def is_continuation(self):
        """ """
        return all([n.is_continuation for n in self.notes])

    @property
    def starts_with_absolute_note(self):
        """ """
        if len(self.notes) > 0:
            return self.notes[0].starts_with_absolute_note
        else:
            return False

    @property
    def had_absolute_note(self):
        """ """
        return any([n.starts_with_absolute_note for n in self.notes])

    @property
    def starts_with_absolute_or_silence(self):
        """ """
        if len(self.notes) > 0:
            return self.notes[0].starts_with_absolute_or_silence
        else:
            return False

    @property
    def starts_with_note(self):
        """ """
        if len(self.notes) > 0:
            return self.notes[0].starts_with_note
        else:
            return False

    def augment(self, value):
        """

        Parameters
        ----------
        value :
            

        Returns
        -------

        """
        return Melody([n.augment(value) for n in self.notes], nb_bars=self.nb_bars, tags=set(self.tags))

    @property
    def duration(self):
        """ """
        return sum([n.duration for n in self.notes])

    def __iter__(self):
        return self.notes.__iter__()

    def __getitem__(self, item):
        return Melody(self.notes[item])

    def __radd__(self, other):
        if other is None:
            return self.copy()

    def __and__(self, other):
        return Melody([n & other for n in self.notes], nb_bars=self.nb_bars, tags=set(self.tags))

    def __matmul__(self, other):
        # Apply a function to each note
        return Melody([n @ other for n in self.notes], nb_bars=self.nb_bars, tags=set(self.tags))

    def __mul__(self, other):
        from .note import Note
        if isinstance(other, int):
            melody_copy = self.copy()
            return Melody(melody_copy.notes * other, nb_bars=self.nb_bars, tags=set(self.tags))
        if isinstance(other, Note):
            return self * Melody([other.copy()], nb_bars=self.nb_bars, tags=set(self.tags))
        else:
            raise Exception('Cannot multiply Melody and ' + str(type(other)))

    def __len__(self):
        return len(self.notes)

    def o(self, octave):
        """

        Parameters
        ----------
        octave :
            

        Returns
        -------

        """
        return Melody([n.o(octave) for n in self.notes], nb_bars=self.nb_bars, tags=set(self.tags))

    def __hasattr__(self, item):
        try:
            self.__getattr__(item)
        except:
            return False
        return True

    def __getattr__(self, item):
        try:
            res = Melody([getattr(n, item) for n in self.notes], nb_bars=self.nb_bars, tags=set(self.tags))
            return res
        except:
            raise AttributeError(f"Not existing property : {item}")

    def copy(self):
        """ """
        return Melody([s.copy() for s in self.notes], tags=set(self.tags), nb_bars=self.nb_bars)

    def __repr__(self):
        return self.to_code()


import pickle
import logging


class ConcatPipeline:
    """
    Pipeline object represents a sequence of steps that will lead to a concatenation of successive transforms.

    For each step of the pipeline : ``score = score + transform(score)``

    Examples
    ---------
    Example applying a transpose only on chords that have an "ok" tag :

    >>> from musiclang.library import *
    >>> from musiclang.transform import ConcatPipeline
    >>> transposer = Transpose(2)
    >>> mask = Mask.Chord() > Mask.Has('ok')

    >>> pipeline = ConcatPipeline([('transpose', transposer, mask)])
    >>> score = (I % I.M)(piano=[s0, s2, s4]).add_tag('ok') + (V % I.M)(piano=[s0, s2, s4])
    >>> new_score = pipeline(score)
    >>> new_score
    (I % I.M)(piano=[s0, s2, s4]) + (V % I.M)(piano=[s0, s2, s4]) + (I % I.M)(piano=[s2, s4, s6]) + (V % I.M)(piano=[s0, s2, s4])

    Same example without the query :

    >>> from musiclang.library import *
    >>> from musiclang.transform import ConcatPipeline
    >>> transposer = Transpose(2)
    >>> pipeline = ConcatPipeline([('transpose', transposer)])
    >>> score = (I % I.M)(piano=[s0, s2, s4]).add_tag('ok') + (V % I.M)(piano=[s0, s2, s4])
    >>> new_score = pipeline(score)
    >>> new_score
    (I % I.M)(piano=[s0, s2, s4]) + (V % I.M)(piano=[s0, s2, s4]) + (I % I.M)(piano=[s2, s4, s6]) + (V % I.M)(piano=[s2, s4, s6])

    """
    def __init__(self, steps):
        self.steps = steps

    @property
    def dict_steps(self):
        return dict(self.steps)

    def save(self, filepath):
        with open(filepath, 'wb') as f:
            pickle.dump(self.steps, f)

    def __repr__(self):
        return "Pipeline({})".format(self.steps.__repr__())

    def get_params(self, **params):
        pass

    def modify(self, param, value):
        candidates = [i for i in range(len(self.steps)) if self.steps[i][0] == param]
        if len(candidates) > 0:
            self.steps[candidates[0]] = (param, value)
        else:
            raise KeyError(f'Not existing param : {param}')

    def set_params(self, _copy=True, **params):
        """
        Set a nested dict of params
        :param params:
        :return:
        """
        if _copy:
            new_pipeline = self.copy()
        else:
            new_pipeline = self
        for param, value in params.items():
            to_modify = new_pipeline.dict_steps[param]
            if isinstance(to_modify, ConcatPipeline):
                new_pipeline.dict_steps[param] = to_modify.set_params(_copy=False, **value)

            else:
                if isinstance(value, dict):
                    new_params = new_pipeline.dict_steps[param].get_params()
                    new_params.update(value)
                    new_pipeline.modify(param, to_modify.__class__(**new_params))
                else:
                    new_pipeline.modify(param, value)

        return new_pipeline

    def copy(self):
        new_pipeline = self.__class__(steps=[])
        for data in self.steps:
            on = Mask()
            if len(data) == 2:
                from .mask import Mask
                step, val = data
            elif len(data) == 3:
                step, val, on = data
            new_pipeline.steps.append((step, val.copy(), on))
        return new_pipeline

    def unpack_args(self, step, f, kwargs):
        if isinstance(f, ConcatPipeline):
            print(kwargs)
            args = {k.split('__')[1]: kwargs[k] for k in kwargs.keys() if k.startswith(step + '__')}
            return args
        else:
            return kwargs

    def __call__(self, score=None, **kwargs):
        from .mask import Mask
        from musiclang import Score
        for data in self.steps:
            on = Mask()
            step, f = '', None
            if len(data) == 2:
                from .mask import Mask
                step, f = data
            elif len(data) == 3:
                step, f, on = data
            arguments = self.unpack_args(step, f, kwargs)
            try:
                to_add = f(score, on=on, **arguments)
                if isinstance(score, Score):
                    score += to_add.add_tag_children(f'step_{step}')
                else:
                    score += to_add
            except Exception as e:
                logging.error(f"Exception in Pipeline: {step}")
                logging.exception(e)
                raise e
        return score

    @staticmethod
    def load(filepath):
        with open(filepath, 'rb') as f:
            return ConcatPipeline(pickle.load(f))


class TransformPipeline(ConcatPipeline):
    """
    Pipeline object represents a sequence of steps that will lead to successive transforms.

    For each step of the pipeline : ``score = transform(score)``


    Examples
    ---------
    Example applying a transpose only on chords that have an "ok" tag :

    >>> from musiclang.library import *
    >>> from musiclang.transform import TransformPipeline
    >>> transposer = Transpose(1)
    >>> mask = Mask.Chord() > Mask.Has('ok')

    >>> pipeline = TransformPipeline([('transpose', transposer, mask), ('transpose2', transposer)])
    >>> score = (I % I.M)(piano=[s0, s2, s4]).add_tag('ok') + (V % I.M)(piano=[s0, s2, s4])
    >>> new_score = pipeline(score)
    >>> new_score
    (I % I.M)(piano=[s2, s4, s6]) + (V % I.M)(piano=[s1, s3, s4])

    The first chord with the tag is transposed twice (first transform match the mask)
    The second one is transposed once

    """

    def __repr__(self):
        return "TransformPipeline({})".format(self.steps.__repr__())

    def __call__(self, score=None, **kwargs):
        from musiclang import Score
        for data in self.steps:
            from .mask import Mask
            on = Mask()
            step, f = None, None
            if len(data) == 2:
                step, f = data
            elif len(data) == 3:
                step, f, on = data

            arguments = self.unpack_args(step, f, kwargs)
            try:
                score = f(score, on=on, **arguments)
                if isinstance(score, Score):
                    score = score.add_tag_children(f'step_{step}')

            except Exception as e:
                logging.error(f"Exception in TransformPipeline step : {step}")
                logging.exception(e)
                raise e

        return score


from .parser import parse_to_musiclang, get_chords_from_analysis, get_chords_from_mxl, \
    parse_midi_to_musiclang, parse_mxl_to_musiclang
from .roman_parser import annotation_to_musiclang
from .score_formatter import ScoreFormatter
from .pattern_sampler import PicklePatternSampler, TextPatternSampler
from .pattern_analyzer import PatternExtractor, PatternFeatureExtractor, recursive_correct_octave, inverse_recursive_correct_octave

__all__ = ['parse_to_musiclang', 'get_chords_from_analysis', 'get_chords_from_mxl',
          'parse_mxl_to_musiclang', 'parse_midi_to_musiclang', 'annotation_to_musiclang', 'ScoreFormatter',
           'PicklePatternSampler', 'TextPatternSampler', 'PatternExtractor', 'PatternFeatureExtractor', 'recursive_correct_octave',
           'inverse_recursive_correct_octave'
           ]


from musiclang import Melody
from musiclang.transform import MelodyTransformer
from musiclang.transform.composing.project import project_on_rhythm


class CircularPermutationMelody(MelodyTransformer):
    """
    Circular permut the melody by a factor n

    Can be called with a melody, a score or a chord

    Examples
    --------

    >>> from musiclang.transform.library import CircularPermutationMelody
    >>> from musiclang.library import *
    >>> melody = s0 + s1 + s2
    >>> CircularPermutationMelody(1)(melody)
    s1 + s2 + s0

    >>> from musiclang.transform.library import CircularPermutationMelody
    >>> from musiclang.library import *
    >>> melody = (I % I.M)(s0 + s1, s1 + s2)
    >>> CircularPermutationMelody(1)(melody)
    (I % I.M)(s1 + s0, s2 + s1)

    """
    def __init__(self, n, **kwargs):
        """

        :param n: Number of delays of
        :param kwargs:
        """
        self.n = n

    def action(self, melody: Melody, **kwargs):
        new_melody = melody.copy()
        notes = [melody.notes[(idx + self.n) % len(melody.notes)] for idx, note in enumerate(melody.notes)]
        new_melody.notes = notes

        return new_melody


class ReverseMelody(MelodyTransformer):
    """
    Reverse the current melody (first note is last etc ...)

    Can be called with a melody, a score or a chord.

    Examples
    --------

    >>> from musiclang.transform.library import ReverseMelody
    >>> from musiclang.library import *
    >>> melody = s0 + s1 + s2
    >>> ReverseMelody(1)(melody)
    s2 + s1 + s0

    >>> from musiclang.transform.library import ReverseMelody
    >>> from musiclang.library import *
    >>> melody = (I % I.M)(s0 + s1, s1 + s2)
    >>> ReverseMelody(1)(melody)
    (I % I.M)(s1 + s0, s2 + s1)


    """

    def action(self, melody, **kwargs):
        return Melody(melody.notes[::-1], tags=melody.tags)


class ReverseMelodyWithoutRhythm(MelodyTransformer):
    """
    Reverse the notes of a melody

    Can be called with a melody, a score or a chord.

    Examples
    --------

    >>> from musiclang.transform.library import ReverseMelodyWithoutRhythm
    >>> from musiclang.library import *
    >>> melody = s0 + s1 + s2.e
    >>> ReverseMelodyWithoutRythm()(melody)
    s2 + s1 + s0.e

    >>> from musiclang.transform.library import ReverseMelodyWithoutRhythm
    >>> from musiclang.library import *
    >>> melody = (I % I.M)(s0 + s1.e, s1 + s2.e)
    >>> ReverseMelodyWithoutRythm()(melody)
    (I % I.M)(s1 + s0.e, s2 + s1.e)
    """

    def action(self, melody, **kwargs):
        return project_on_rhythm(melody, Melody(melody.notes[::-1], tags=melody.tags))


class InvertMelody(MelodyTransformer):
    """
    Invert the melody from the first note of the melody

    Can be called with a melody, a score or a chord.

    Examples
    --------

    >>> from musiclang.transform.library import InvertMelody
    >>> from musiclang.library import *
    >>> melody = s0 + s1 + s2.e
    >>> InvertMelody()(melody)
    s2 + s6.o(-1) + s5.o(-1).e

    >>> from musiclang.transform.library import InvertMelody
    >>> from musiclang.library import *
    >>> melody = (I % I.M)(s0 + s1.e, s1 + s2.e)
    >>> InvertMelody()(melody)
    (I % I.M)(s0 + s6.o(-1).e, s1 + s0.e)
    """

    def action(self, melody, **kwargs):
        first_val = melody.notes[0].val
        return Melody([n.set_val(2 * first_val - n.val) for n in melody.notes], tags=melody.tags)


class SelectRangeMelody(MelodyTransformer):
    """
    Get the sub-melody from time start and time end.

    Can be called with a melody, a chord or a score (in which case it will be applied chord per chord).

    Examples
    --------

    >>> from musiclang.transform.library import SelectRangeMelody
    >>> from musiclang.library import *
    >>> melody = s0 + s1 + s2.e
    >>> SelectRangeMelody(0, 1)(melody)
    s0

    >>> from musiclang.transform.library import SelectRangeMelody
    >>> from musiclang.library import *
    >>> melody = (I % I.M)(s0 + s1.e + s2) + (I % I.M)(s0 + s1.e + s2)
    >>> SelectRangeMelody(0, 2)(melody)
    (I % I.M)(piano__0=s0 + s1.e + s2.e) + (I % I.M)(piano__0=s0 + s1.e + s2.e)
    """
    def __init__(self, start=None, end=None):
        self.start = start
        self.end = end

    def action(self, melody, **kwargs):
        from musiclang.write.time_utils import get_score_between
        from musiclang.library import I
        # Get melody between ...
        return get_score_between(None + (I % I.M)(piano__0=melody), start=self.start, end=self.end)[0].score['piano__0']





from musiclang.transform import NoteTransformer


class LimitRegister(NoteTransformer):
    """
    Limit octave and values of notes (In tonal values). Transpose the octave until the note fit the range

    Can be called with a note, a melody, a score or a chord.
    It will be applied for each note.

    Examples
    --------

    >>> from musiclang.transform.library import LimitRegister
    >>> from musiclang.library import *
    >>> melody = s0 + s1.o(1)
    >>> LimitRangeMelody(s0, s0.o(1))(melody)
    s0 + s1
    """
    def __init__(self, note_min, note_max):
        self.note_min = note_min
        self.note_max = note_max

        if self.pitch_max - self.pitch_min < 7:
            raise ValueError('There should be a span of at least one octave for the limiter')

    @property
    def pitch_min(self):
        return self.note_min.scale_pitch

    @property
    def pitch_max(self):
        return self.note_max.scale_pitch

    def limit(self, note):
        sp = note.scale_pitch
        if sp > self.pitch_max:
            return self.limit(note.o(-1))
        elif sp < self.pitch_min:
            return self.limit(note.o(1))
        else:
            return note.copy()

    def action(self, note, **kwargs):
        return self.limit(note)


class ApplySilence(NoteTransformer):
    """
    Transform all notes into a Silence

    """
    def action(self, element, **kwargs):
        from musiclang import Silence
        return Silence(element.duration)


class ApplyContinuation(NoteTransformer):
    def action(self, element, **kwargs):
        from musiclang import Continuation
        return Continuation(element.duration)


class Repeat(NoteTransformer):
    """
    Transform all notes into a Continuation
    """
    def __init__(self, n):
        self.n = n

    def action(self, note, **kwargs):
        return note * self.n


class TransposeDiatonic(NoteTransformer):
    """
    Transpose a note diatonically
    """
    def __init__(self, n, keep_mode=True, keep_accident=False):
        """

        Parameters
        ----------
        n: int
           Number of diatonic steps for transposition
        keep_mode: boolean, default=True
                   If false the transposed notes will lose it's mode
        keep_accident: boolean, default=False
                   If false the transposed notes will lose it's accident
        """
        self.n = n
        self.keep_mode = keep_mode
        self.keep_accident = keep_accident

    def action(self, note, chord=None, **kwargs):
        if note.type in ['s']:
            new_note = note.copy()
            new_note.type = 's'
            new_note.val += self.n
            if not self.keep_mode:
                new_note.mode = None
            elif not self.keep_accident:
                new_note.accident = None
            return new_note
        elif note.is_relative:
            return note.copy()

        elif note.type == 'h':
            from musiclang.write.constants import DEGREE_TO_SCALE_DEGREE
            new_val = DEGREE_TO_SCALE_DEGREE[note.val]
            new_note = note.copy()
            new_note.type = 's'
            new_note.val = new_val
            return new_note


class TransposeChromatic(NoteTransformer):
    """
    Transpose a note chromatically.
    Should be called on at least a chord
    """
    def __init__(self, n):
        self.n = n

    def action(self, note, chord=None, **kwargs):
        if note.is_relative:
            return note.copy()

        dict_pitches = chord.pitch_dict
        pitch = chord.to_pitch(note)
        ref_note = dict_pitches[(pitch + self.n) % 12].o((pitch + self.n)//12)
        new_note = note.copy()
        new_note.val = ref_note.val
        new_note.type = ref_note.type
        new_note.octave = ref_note.octave
        new_note.accident = None
        new_note.mode = None

        return new_note


"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

import numpy as np
def relative_scale_up_value(delta, last_pitch, scale_pitches):
    """

    Parameters
    ----------
    note : Note
    last_pitch : int
    scale_pitches : list[int]


    Returns
    -------
    pitch: int

    """
    scale_mod = list(sorted([i % 12 for i in scale_pitches]))
    # Get scale index
    if delta == 0:
        whole_scale = [s + octave * 12 for octave in range(-10, 10) for s in scale_mod]
        whole_scale = [s for s in whole_scale if s >= last_pitch]
        new_pitch = whole_scale[0]
        return new_pitch
    # else:
    whole_scale = [s + octave * 12 for octave in range(-10, 10) for s in scale_mod]
    dists = np.asarray(whole_scale) - last_pitch
    dists = dists[dists >= 0] + last_pitch
    new_pitch = dists[delta - 1 * (last_pitch not in whole_scale)]
    return new_pitch


def relative_scale_down_value(delta, last_pitch, scale_pitches):
    """

    Parameters
    ----------
    note : Note
    last_pitch : int
    scale_pitches : list[int]


    Returns
    -------
    pitch: int

    """
    scale_mod = list(sorted([i % 12 for i in scale_pitches]))
    # Get scale index
    if delta == 0:
        whole_scale = [s + octave * 12 for octave in range(-10, 10) for s in scale_mod]
        whole_scale = [s for s in whole_scale if s <= last_pitch]
        new_pitch = whole_scale[-1]
        return new_pitch
    # else:
    whole_scale = [s + octave * 12 for octave in range(-10, 10) for s in scale_mod]
    dists = np.asarray(whole_scale) - last_pitch
    dists = dists[dists <= 0] + last_pitch
    new_pitch = dists[-(delta + 1) + 1 * (last_pitch not in whole_scale)]
    return new_pitch


def get_relative_scale_value(note, last_pitch, scale_pitches):
    """
    Parameters
    ----------
    note : Note
    last_pitch : int
    scale_pitches : list[int]

    Returns
    -------
    pitch: int
    """
    # Remove non unique scale_pitches
    scale_pitches = list(sorted(set([s % 12 for s in scale_pitches])))
    total_val = note.val + len(scale_pitches) * note.octave
    if note.is_down: # su or sd
        total_val = - total_val
    if total_val > 0:
        return relative_scale_up_value(total_val, last_pitch, scale_pitches)
    elif total_val < 0:
        return relative_scale_down_value(-total_val, last_pitch, scale_pitches)
    elif total_val == 0:
        if last_pitch % 12 in [s % 12 for s in scale_pitches]:
            return last_pitch
        pitch_up = relative_scale_up_value(total_val, last_pitch, scale_pitches)
        pitch_down = relative_scale_down_value(-total_val, last_pitch, scale_pitches)
        # Select nearest from last pitch
        if abs(pitch_up - last_pitch) <= abs(pitch_down - last_pitch):
            return pitch_up
        return pitch_down
    else:
        raise Exception(f'This relative note {note} is nor up or down')


def get_value_to_scale_note_with_accident(note, chord):
    """
    Get the value of the note if there is an accident
    Parameters
    ----------
    note: Note
    chord: Chord

    Returns
    -------
    pitch: int

    """
    from musiclang.write.constants import ACCIDENTS_TO_NOTE
    tonic = chord.scale_pitches[0]
    return tonic + ACCIDENTS_TO_NOTE[(note.val, note.accident)] + 12 * (note.octave)

def get_value_to_scale_note(value, scale_pitches):
    """
    Transform an absolute value and a scale pitches to an unique pitch

    Parameters
    ----------
    value : int
        
    scale_pitches : list[int]
        

    Returns
    -------
    pitch: int

    """
    return scale_pitches[value % len(scale_pitches)] + 12 * (value // len(scale_pitches))

def note_to_pitch_result(note, chord, last_pitch=None):
    """

    Parameters
    ----------
    note :
        
    chord :
        
    last_pitch :
         (Default value = None)

    Returns
    -------
    pitch: int
           Resulting pitch

    """
    real_chord = note.real_chord(chord)
    scale_pitches = real_chord.scale_pitches

    pitch_result = None
    if not note.is_relative:
        if note.is_scale_note:
            if note.has_accident:
                pitch_result = get_value_to_scale_note_with_accident(note, real_chord)
            else:
                pitch_result = get_value_to_scale_note(note.val + 7 * note.octave, scale_pitches)
        elif note.is_chord_note:
            scale_chord = chord.chord_pitches
            pitch_result = get_value_to_scale_note(note.val + len(scale_chord) * note.octave, scale_chord)
        elif note.is_bass_note:
            scale_chord = chord.chord_extension_pitches
            pitch_result = get_value_to_scale_note(note.val + len(scale_chord) * note.octave, scale_chord)
        elif note.is_absolute_note:
            chromatic_pitches = list(range(12))
            pitch_result = get_value_to_scale_note(note.val + 12 * note.octave, chromatic_pitches)
        elif note.is_chromatic_note:
            chromatic_pitches = [scale_pitches[0] + i for i in range(12)]
            pitch_result = get_value_to_scale_note(note.val + 12 * note.octave, chromatic_pitches)
        elif note.is_drum_note:
            chromatic_pitches = list(range(12))
            pitch_result = get_value_to_scale_note(note.val + 12 * note.octave, chromatic_pitches)

    elif note.is_relative:
        if note.is_scale_note:
            pitch_result = get_relative_scale_value(note, last_pitch, scale_pitches)
            pass
        elif note.is_chord_note:
            scale_chord = chord.chord_pitches
            pitch_result = get_relative_scale_value(note, last_pitch, scale_chord)
            pass
        elif note.is_bass_note:
            scale_chord = chord.chord_extension_pitches
            pitch_result = get_relative_scale_value(note, last_pitch, scale_chord)
            pass
        elif note.is_chromatic_note:
            chromatic_pitches = chord.chromatic_pitches
            pitch_result = get_relative_scale_value(note, last_pitch, chromatic_pitches)
            pass
        else:
            raise Exception('This kind of note is not supported or has no pitch : {}'.format(note.type))

    return pitch_result



from .element import Formatter

class TimeSignatureElement(Formatter):

    def __init__(self, num, den):
        self.signature = (num, den)

    def to_text(self):
        return '\n' + str(self)

    def __repr__(self):
        return f'Time Signature:{self.signature[0]}/{self.signature[1]}'



"""Classes and data structures related to tonal features."""

import numpy as np

from .chord_vocabulary import frompcset

NOTENAMES = ("C", "D", "E", "F", "G", "A", "B")

NOTENAMES_LOWERCASE = [n.lower() for n in NOTENAMES]

PITCHCLASSES = list(range(12))

ACCIDENTALS = ("--", "-", "", "#", "##")

SPELLINGS = [
    f"{letter}{accidental}"
    for letter in NOTENAMES
    for accidental in ACCIDENTALS
]

INTERVALCLASSES = [
    f"{specific}{generic}"
    for generic in [2, 3, 6, 7]
    for specific in ["dd", "d", "m", "M", "A", "AA"]
] + [
    f"{specific}{generic}"
    for generic in [1, 4, 5]
    for specific in ["dd", "d", "P", "A", "AA"]
]

DEGREES = (
    "-1",
    "-2",
    "-3",
    "-4",
    "-5",
    "-6",
    "-7",
    "1",
    "2",
    "3",
    "4",
    "5",
    "6",
    "7",
    "#1",
    "#2",
    "#3",
    "#4",
    "#5",
    "#6",
    "#7",
    "None",
)

KEYS = tuple(sorted(set([key for keys in frompcset.values() for key in keys])))

# The keys used for transposition (data augmentation)
TRANSPOSITIONKEYS = tuple(KEYS)

CHORD_QUALITIES = tuple(
    sorted(
        set(
            [
                key["quality"]
                for keys in frompcset.values()
                for key in keys.values()
            ]
        )
    )
)

COMMON_ROMAN_NUMERALS = tuple(
    [
        # Cadentials are undistinguishable from a I chord in the vocabulary,
        # they are contextually (and explicitly) annotated by the analyst
        "Cad"
    ]
    + sorted(
        set(
            [key["rn"] for keys in frompcset.values() for key in keys.values()]
        )
    )
)

PCSETS = tuple(sorted(frompcset.keys()))

INTERVAL_ENHARMONICS = {
    "A1": "m2",
    "M2": "D3",
    "A2": "m3",
    "M3": "D4",
    "A3": "P4",
    "A4": "D5",
    "P5": "D6",
    "A5": "m6",
    "M6": "D7",
    "A6": "m7",
    "M7": "D8",
    "m2": "A1",
    "D3": "M2",
    "m3": "A2",
    "D4": "M3",
    "P4": "A3",
    "D5": "A4",
    "D6": "P5",
    "m6": "A5",
    "D7": "M6",
    "m7": "A6",
    "D8": "M7",
}

NOTEDURATIONS = [
    0,  # onset
    1,  # thirtysecond
    2,  # sixteenth
    3,  # eighth
    4,  # quarter
    5,  # half
    6,  # whole
]


class FeatureRepresentation(object):
    """ """
    features = 1

    def __init__(self, df):
        self.df = df
        self.frames = len(df.index)
        self.dtype = "i8"
        self.array = self.run()

    @property
    def shape(self):
        """ """
        return (self.frames, self.features)

    def run(self, tranposition=None):
        """

        Parameters
        ----------
        tranposition :
             (Default value = None)

        Returns
        -------

        """
        array = np.zeros(self.shape, dtype=self.dtype)
        return array

    def dataAugmentation(self, intervals):
        """

        Parameters
        ----------
        intervals :
            

        Returns
        -------

        """
        for interval in intervals:
            yield self.run(transposition=interval)
        return

    @classmethod
    def encodeManyHot(cls, array, timestep, index, value=1):
        """

        Parameters
        ----------
        array :
            
        timestep :
            
        index :
            
        value :
             (Default value = 1)

        Returns
        -------

        """
        if 0 <= index < cls.features:
            array[timestep, index] = value
        else:
            raise IndexError

    @classmethod
    def encodeCategorical(cls, array, timestep, classNumber):
        """

        Parameters
        ----------
        array :
            
        timestep :
            
        classNumber :
            

        Returns
        -------

        """
        if 0 <= classNumber < cls.features:
            array[timestep] = classNumber
        else:
            raise IndexError


class FeatureRepresentationTI(FeatureRepresentation):
    """TI stands for Transposition Invariant.
    
    If a representation is TI, dataAugmentation consists of

    Parameters
    ----------

    Returns
    -------
    type
        

    """

    def dataAugmentation(self, intervals):
        """

        Parameters
        ----------
        intervals :
            

        Returns
        -------

        """
        for _ in intervals:
            yield np.copy(self.array)
        return



from .element import Formatter

class CounterpointElement(Formatter):

    def __init__(self, fixed_voices=None):
        self.fixed_voices = fixed_voices

    def to_text(self):
        return '\n' + str(self)

    def __repr__(self):
        return f'!counterpoint {" ".join(self.fixed_voices)}'


"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

import numpy as np

class SkylineSolution:
    """ """
    def get_skyline(self, notes):
        """

        Parameters
        ----------
        notes :
            

        Returns
        -------

        """
        if not notes:
            return []
        if len(notes) == 1:
            return [[notes[0][0], notes[0][2]] + notes[0][3:], [notes[0][1], 0] + notes[0][3:]]

        mid = len(notes) // 2
        left = self.get_skyline(notes[:mid])
        right = self.get_skyline(notes[mid:])
        return self.merge(left, right)

    def merge(self, left, right):
        """

        Parameters
        ----------
        left :
            
        right :
            

        Returns
        -------

        """
        h1, h2 = 0, 0
        i, j = 0, 0
        note1, note2 = [], []
        result = []

        while i < len(left) and j < len(right):
            if left[i][0] < right[j][0]:
                h1 = left[i][1]
                note1 = left[i][2:]
                corner = left[i][0]
                i += 1
            elif right[j][0] < left[i][0]:
                h2 = right[j][1]
                note2 = right[j][2:]
                corner = right[j][0]
                j += 1
            else:
                h1 = left[i][1]
                note1 = left[i][2:]
                h2 = right[j][1]
                note2 = right[j][2:]
                corner = right[j][0]
                i += 1
                j += 1
            if self.is_valid(result, max(h1, h2)):
                max_note = [note1, note2][np.argmax([h1, h2])]
                result.append([corner, max(h1, h2)] + max_note)
        result.extend(right[j:])
        result.extend(left[i:])
        return result

    def is_valid(self, result, new_height):
        """

        Parameters
        ----------
        result :
            
        new_height :
            

        Returns
        -------

        """
        return not result or result[-1][1] != new_height



from transformers import Trainer, TrainingArguments, AutoModelWithLMHead, GPT2LMHeadModel, GPT2Config, \
    GPT2TokenizerFast
from transformers import TextDataset,DataCollatorForLanguageModeling
from transformers import AutoTokenizer, GPT2TokenizerFast
import os
import glob
from tokenizers import ByteLevelBPETokenizer


def load_dataset(train_path, val_path, tokenizer, block_size=1024):
    train_dataset = TextDataset(
        tokenizer=tokenizer,
        file_path=train_path,
        block_size=block_size)

    val_dataset = TextDataset(
        tokenizer=tokenizer,
        file_path=val_path,
        block_size=block_size)

    data_collator = DataCollatorForLanguageModeling(
        tokenizer=tokenizer, mlm=False,
    )
    return train_dataset, val_dataset, data_collator


def train_custom_tokenizer(input_file, output_directory, vocab_size=200, min_frequency=2):
    tokenizer = ByteLevelBPETokenizer()
    tokenizer.train(
        files=[input_file],
        vocab_size=vocab_size,
        min_frequency=min_frequency,
        special_tokens=["<s>", "</s>", "<unk>", "<mask>"],
    )
    os.makedirs(output_directory, exist_ok=True)
    tokenizer.save_model(output_directory)
    return tokenizer


class DatasetFinetuner:
    """
    Main class for training a musiclang LLM model from an existing model

    Examples
    --------
    Finetune existing musiclangLLM
    >>>from musiclang.analyze.dataset_trainer import DatasetFinetuner
    >>>from transformers import GPT2TokenizerFast, GPT2LMHeadModel
    >>>model_hub_name = "floriangardin/musiclang"
    >>>train_dataset = "locals/data/training/train.txt"
    >>>val_dataset = "locals/data/training/val.txt"
    >>>output_dir = "locals/data/training/model"

    >>>model = GPT2LMHeadModel.from_pretrained(model_hub_name)
    >>>tokenizer = GPT2TokenizerFast.from_pretrained(model_hub_name)
    >>>finetuner = DatasetFinetuner(model, tokenizer, train_dataset, val_dataset, output_dir)
    >>>model, tokenizer = finetuner.train()
    """
    def __init__(self, model, tokenizer, train_dataset, val_dataset, output_dir):
        """

        Parameters
        ----------
        model: HuggingFace model
        tokenizer: HuggingFace tokenizer
        train_dataset: str
            Path of the concatenated train file
        val_dataset: str
            Path of the concatenated test file
        output_dir: str
            Where to save the model
        """

        self.model = model
        self.tokenizer = tokenizer
        self.output_dir = output_dir
        self.train_dataset = train_dataset
        self.val_dataset = val_dataset

    def train(self, epochs=5, batch_size=16, warmup_steps=500, block_size=1024, eval_steps=800):

        train_dataset, val_dataset, data_collator = load_dataset(self.train_dataset, self.val_dataset,
                                                                 self.tokenizer, block_size)

        training_args = TrainingArguments(
            output_dir=self.output_dir,  # The output directory
            overwrite_output_dir=True,  # overwrite the content of the output directory
            num_train_epochs=epochs,  # number of training epochs
            per_device_train_batch_size=batch_size,  # batch size for training
            per_device_eval_batch_size=batch_size,  # batch size for evaluation
            eval_steps=eval_steps,  # Number of update steps between two evaluations.
            save_steps=eval_steps,  # after # steps model is saved
            warmup_steps=warmup_steps,  # number of warmup steps for learning rate scheduler
            prediction_loss_only=True,
            evaluation_strategy="epoch",
            do_eval=True
        )

        trainer = Trainer(
            model=self.model,
            args=training_args,
            data_collator=data_collator,
            train_dataset=train_dataset,
            eval_dataset=val_dataset,
        )
        print('Will train the model')
        trainer.train()
        trainer.save_model()

        return self.model, self.tokenizer


class DatasetTrainer:
    """
    Train a LLM from scratch using a configuration
    """

    def __init__(self, config_model, config_tokenizer, train_dataset, val_dataset, output_dir):
        self.config_model = config_model
        self.output_dir = output_dir
        self.train_dataset = train_dataset
        self.config_tokenizer = config_tokenizer
        self.val_dataset = val_dataset
        self.model = None
        self.tokenizer = None

    def preprocess(self):
        tokenizer = train_custom_tokenizer(self.train_dataset, self.output_dir, **self.config_tokenizer)
        return tokenizer
    def train(self, epochs=5, batch_size=16, warmup_steps=500, block_size=1024, eval_steps=800):
        # Train the tokenizer
        self.tokenizer = self.preprocess()
        train_dataset, val_dataset, data_collator = load_dataset(self.train_dataset, self.val_dataset,
                                                                 self.tokenizer, block_size)

        # Train the model

        config = GPT2Config(vocab_size=len(self.tokenizer),
                            bos_token_id=self.tokenizer.bos_token_id,
                            eos_token_id=self.tokenizer.eos_token_id,
                            **self.config_model
                            )

        # Initialize a new GPT-2 model with the custom configuration
        self.model = GPT2LMHeadModel(config)

        training_args = TrainingArguments(
            output_dir=self.output_dir,  # The output directory
            overwrite_output_dir=True,  # overwrite the content of the output directory
            num_train_epochs=epochs,  # number of training epochs
            per_device_train_batch_size=batch_size,  # batch size for training
            per_device_eval_batch_size=batch_size,  # batch size for evaluation
            eval_steps=eval_steps,  # Number of update steps between two evaluations.
            save_steps=eval_steps,  # after # steps model is saved
            warmup_steps=warmup_steps,  # number of warmup steps for learning rate scheduler
            prediction_loss_only=True,
            evaluation_strategy="epoch",
            do_eval=True
        )

        trainer = Trainer(
            model=self.model,
            args=training_args,
            data_collator=data_collator,
            train_dataset=train_dataset,
            eval_dataset=val_dataset,
        )

        trainer.train()
        trainer.save_model()




"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

from .constants import SCALE_DEGREE

class Element:
    """
    An element represents the basis of numeral roman notation, it usually represents a degree.

    In the library elements are represented with roman notation ``I,II,III,IV,V,VI,VII``
    The first degree is indexed with a 0 value (I has a value of 0)
    It is an intermediary object destined to create chords or tonalities :
    - If you call or modulate this object you get a Chord
    - if you call a mode ``m,M,mm,dorian,phrygian,lydian,mixolydian,aeolian,locrian`` on this you get a Tonality

    Parameters
    ----------

    val: int,
         Degree
    """
    def __init__(self, val):
        self.val = val


    def __add__(self, other):
        from musiclang import Chord, Score
        if isinstance(other, int):
            return Element(self.val + other)
        if isinstance(other, Element):
            return self() + other()
        elif isinstance(Element, (Chord, Score)):
            return self() + other
        else:
            raise Exception('Cannot add if not type int or Element')

    def __repr__(self):
        from .chord import Chord
        return Chord(element=self.val).__repr__()

    def __getitem__(self, extension):
        from .chord import Chord
        return Chord(element=self.val)[extension]

    def __call__(self, *melodies, **named_melodies):
        from .chord import Chord
        return Chord(element=self.val)(*melodies, **named_melodies)


    def __mod__(self, other):
            from .chord import Chord
            return Chord(element=self.val) % other


    def copy(self):
        """ """
        return Element(self.val)

    def o(self, octave):
        """
        Create a new chord from this element octaved with given parameter
        Parameters
        ----------
        octave :
            

        Returns
        -------
        chord: Chord
               The result chord transpose by n octaves
        """
        from .chord import Chord
        return Chord(element=self.val, extension=5).o(octave)


    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__ = d

    @property
    def pitch_set(self):
        """
        Returns chords.pitch_set

        Returns
        -------
         pitches: set
                  The pitches associated with this element
        """
        return self.o(0).pitch_set

    @property
    def scale_set(self):
        """ """
        return self.o(0).scale_set

    @property
    def scale_pitches(self):
        """ """
        return self.o(0).scale_pitches

    @property
    def element(self):
        """ """
        return self.val

    @property
    def tonality(self):
        """ """
        from .tonality import Tonality
        return Tonality(0)

    @property
    def octave(self):
        """ """
        return 0

    @property
    def b(self):
        """ """
        from .tonality import Tonality
        return Tonality(SCALE_DEGREE[self.val]).b

    @property
    def s(self):
        """ """
        from .tonality import Tonality
        return Tonality(SCALE_DEGREE[self.val]).s

    @property
    def m(self):
        """ """
        from .tonality import Tonality
        return Tonality(SCALE_DEGREE[self.val]).m

    @property
    def M(self):
        """ """
        from .tonality import Tonality
        return Tonality(SCALE_DEGREE[self.val]).M

    @property
    def mm(self):
        """ """
        from .tonality import Tonality
        return Tonality(SCALE_DEGREE[self.val]).mm


    @property
    def dorian(self):
        """ """
        from .tonality import Tonality
        return Tonality(SCALE_DEGREE[self.val]).dorian

    @property
    def phrygian(self):
        """ """
        from .tonality import Tonality
        return Tonality(SCALE_DEGREE[self.val]).phrygian

    @property
    def lydian(self):
        """ """
        from .tonality import Tonality
        return Tonality(SCALE_DEGREE[self.val]).lydian

    @property
    def mixolydian(self):
        """ """
        from .tonality import Tonality
        return Tonality(SCALE_DEGREE[self.val]).mixolydian

    @property
    def aeolian(self):
        """ """
        from .tonality import Tonality
        return Tonality(SCALE_DEGREE[self.val]).aeolian

    @property
    def locrian(self):
        """ """
        from .tonality import Tonality
        return Tonality(SCALE_DEGREE[self.val]).locrian




"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

import numpy as np
import pandas as pd

from .skyline_algorithm import SkylineSolution


def reduce_one(sequence, high=True):
    """

    Parameters
    ----------
    sequence :
        
    high :
         (Default value = True)

    Returns
    -------

    """
    sequence = sequence.copy()
    SILENCE_ABS_VAL = 2000
    silence_value = -SILENCE_ABS_VAL if high else SILENCE_ABS_VAL
    # Use skyline algorithm to find height map
    notes = sequence[['start', 'end', 'pitch', 'note_type', 'note_val', 'note_octave', 'note_duration',
                      'note_amp', 'chord_degree', 'chord_extension', 'chord_octave', 'tonality_degree',
                      'tonality_mode', 'tonality_octave', 'chord_idx', 'instrument', 'note_idx']].fillna(
        silence_value)

    if not high:
        notes['pitch'] = - notes['pitch']
    notes['pitch'] += SILENCE_ABS_VAL

    # Add little randomness to store equals pitches
    random_noise = np.random.random(len(notes)) / 3
    notes['pitch'] += random_noise
    notes = notes.values.tolist()
    solution = pd.DataFrame(SkylineSolution().get_skyline(notes),
                            columns=['start', 'pitch', 'note_type', 'note_val', 'note_octave', 'note_duration',
                      'note_amp', 'chord_degree', 'chord_extension', 'chord_octave', 'tonality_degree',
                      'tonality_mode', 'tonality_octave', 'chord_idx', 'instrument', 'note_idx'])
    solution['end'] = solution['start'] + solution['start'].diff(1).shift(-1)
    solution['pitch'] -= SILENCE_ABS_VAL
    if not high:
        solution['pitch'] = - solution['pitch']
    solution.loc[solution['pitch'] == silence_value, 'pitch'] = np.nan
    solution = solution.iloc[:-1]
    return solution


def recalculate_pitch(row):
    """

    Parameters
    ----------
    row :
        

    Returns
    -------

    """
    from musiclang.write.chord import Chord
    from musiclang.write.tonality import Tonality
    from musiclang.write.note import Note
    tonality = Tonality(row['tonality_degree'], mode=row['tonality_mode'], octave=row['tonality_octave'])
    chord = Chord(row['chord_degree'], octave=row['chord_octave'], tonality=tonality)
    note = Note(row['note_type'], row['note_val'], row['note_octave'], row['note_duration'], amp=row['note_amp'])
    return chord.to_pitch(note)

def reduce(score, n_voices=4, start_low=False, instruments=None):
    """Arrange a score to sum it up of 4 voices

    Parameters
    ----------
    score :
        param n_voices:
    n_voices :
         (Default value = 4)
    start_low :
         (Default value = False)
    instruments :
         (Default value = None)

    Returns
    -------

    """

    from ..note import Silence
    sequence = score.to_sequence()

    # Use skyline algorithm to find enveloppes
    if instruments is None:
        instruments = [f'piano__{idx}' for idx in range(n_voices)]
    if not start_low:
        instruments = instruments[::-1]
    idxs = [0, -1, 1, -2, 2, -3, 3, -4, 4, -5, 5, -6, 6][:n_voices]
    solutions = []
    for i, idx in enumerate(idxs):
        solution = reduce_one(sequence, high=(i % 2) == 1 * start_low)
        solution['instrument'] = instruments[idx]
        solutions.append(solution)
        # Replace all solution notes in sequence with silences
        replace_silence = set(solution['note_idx'].unique())
        indexer = sequence['note_idx'].isin(replace_silence)
        len_indexer = indexer.sum()
        sequence.loc[indexer, 'note_type'] = ['r' for i in range(len_indexer)]
        sequence.loc[indexer, 'pitch'] = np.nan

    solution = pd.concat(solutions, axis=0).sort_values('start')

    # For each voice reassign to most probable pitch average
    # Recalculate pitch
    solution['pitch'] = solution.apply(recalculate_pitch, axis=1)
    # Recalculate is_silence
    # For each chord remove voices with only silences
    # For each chord each voice calculate average pitch
    # Then starting with first chord reassign each voice with closenessness of average pitch

    # Remove silences

    # Parse solution to new sequence
    from ..score import Score
    solution['silence'] = solution['note_type'] == 'r'
    solution['continuation'] = solution['note_type'] == 'l'
    solution['note_duration'] = solution['end'] - solution['start']
    score_result = Score.from_sequence(solution)

    return score_result





"""Turns a MusicXML file into a pandas DataFrame."""

import music21
from music21.note import Rest
import numpy as np
import pandas as pd

from .cache import m21Interval

FLOATSCALE = 4

# Sixteenth notes
FRAMEBASENOTE = 32
FIXEDOFFSET = round(4.0 / FRAMEBASENOTE, FLOATSCALE)

S_COLUMNS = [
    "s_offset",
    "s_duration",
    "s_measure",
    "s_notes",
    "s_intervals",
    "s_isOnset",
]

S_LISTTYPE_COLUMNS = [
    "s_notes",
    "s_intervals",
    "s_isOnset",
]


def m21Parse(f, fmt=None, remove_perc=True):
    """

    Parameters
    ----------
    f :

    fmt :
         (Default value = None)

    Returns
    -------

    """
    s = music21.converter.parse(f, format=fmt, forceSource=True)
    if remove_perc:

        for el in s.recurse():
            if set(el.classes).intersection(['PercussionChord', 'Unpitched']):  # or 'Piano'
                # el.activeSite.replace(el, instrument.Violin())
                el.activeSite.remove(el)

        # perc = [
        #     p
        #     for p in s.parts
        #     if list(p.recurse().getElementsByClass(["PercussionClef", "PercussionChord"]))
        # ]
        # s.remove(perc, recurse=True)
    return s


def from_tsv(tsv, sep="\t"):
    """

    Parameters
    ----------
    tsv :
        
    sep :
         (Default value = "\t")

    Returns
    -------

    """
    df = pd.read_csv(tsv, sep=sep)
    df.set_index("s_offset", inplace=True)
    for col in S_LISTTYPE_COLUMNS:
        df[col] = df[col].apply(eval)
    return df


def _measureNumberShift(m21Score):
    """

    Parameters
    ----------
    m21Score :
        

    Returns
    -------

    """
    firstMeasure = m21Score.parts[0].measure(0) or m21Score.parts[0].measure(1)
    isAnacrusis = True if firstMeasure.paddingLeft > 0.0 else False
    if isAnacrusis and firstMeasure.number == 1:
        measureNumberShift = -1
    else:
        measureNumberShift = 0
    return measureNumberShift


def _lastOffset(m21Score):
    """

    Parameters
    ----------
    m21Score :
        

    Returns
    -------

    """
    lastMeasure = m21Score.parts[0].measure(-1)
    filledDuration = lastMeasure.duration.quarterLength / float(
        lastMeasure.barDurationProportion()
    )
    lastOffset = lastMeasure.offset + filledDuration
    return lastOffset


def _initialDataFrame(s, fmt=None):
    """Parses a score and produces a pandas dataframe.
    
    The features obtained are the note names, their position in the score,
    measure number, and their ties (in case something fancy needs to be done,
    with the tie information).

    Parameters
    ----------
    s :
        
    fmt :
         (Default value = None)

    Returns
    -------

    """
    dfdict = {col: [] for col in S_COLUMNS}
    measureNumberShift = _measureNumberShift(s)
    for c in s.chordify().flat.notesAndRests:
        dfdict["s_offset"].append(round(float(c.offset), FLOATSCALE))
        dfdict["s_duration"].append(round(float(c.quarterLength), FLOATSCALE))
        dfdict["s_measure"].append(c.measureNumber + measureNumberShift)
        if isinstance(c, Rest):
            # We need dummy entries for rests at the beginning of a measure
            dfdict["s_notes"].append(np.nan)
            dfdict["s_intervals"].append(np.nan)
            dfdict["s_isOnset"].append(np.nan)
            continue
        dfdict["s_notes"].append([n.pitch.nameWithOctave for n in c])
        pitches = [p.nameWithOctave for p in c.pitches]
        intervs = [m21Interval(pitches[0], p).simpleName for p in pitches[1:]]
        dfdict["s_intervals"].append(intervs)
        onsets = [(not n.tie or n.tie.type == "start") for n in c]
        dfdict["s_isOnset"].append(onsets)
    df = pd.DataFrame(dfdict)
    currentLastOffset = float(df.tail(1).s_offset) + float(
        df.tail(1).s_duration
    )
    deltaDuration = _lastOffset(s) - currentLastOffset
    df.loc[len(df) - 1, "s_duration"] += deltaDuration
    df.set_index("s_offset", inplace=True)
    df = df[~df.index.duplicated()]

    return df


def _reindexDataFrame(df, fixedOffset=FIXEDOFFSET):
    """Reindexes a dataframe according to a fixed note-value.
    
    It could be said that the DataFrame produced by parseScore
    is a "salami-sliced" version of the score. This is intuitive
    for humans, but does not really work in machine learning.
    
    What works, is to slice the score in fixed note intervals,
    for example, a sixteenth note. This reindex function does
    exactly that.

    Parameters
    ----------
    df :
        
    fixedOffset :
         (Default value = FIXEDOFFSET)

    Returns
    -------

    """
    firstRow = df.head(1)
    lastRow = df.tail(1)
    minOffset = firstRow.index.to_numpy()[0]
    maxOffset = (lastRow.index + lastRow.s_duration).to_numpy()[0]
    newIndex = np.arange(minOffset, maxOffset, fixedOffset)
    # All operations done over the full index, i.e., fixed-timesteps
    # plus original onsets. Later, original onsets (e.g., triplets)
    # are removed and just the fixed-timesteps are kept
    df = df.reindex(index=df.index.union(newIndex))
    df.s_notes.fillna(method="ffill", inplace=True)
    df.s_notes.fillna(method="bfill", inplace=True)
    # the "isOnset" column is hard to generate in fixed-timesteps
    # however, it allows us to encode a "hold" symbol if we wanted to
    newCol = pd.Series(
        [[False] * n for n in df.s_notes.str.len().to_list()], index=df.index
    )
    df.s_isOnset.fillna(value=newCol, inplace=True)
    df.fillna(method="ffill", inplace=True)
    df.fillna(method="bfill", inplace=True)
    df = df.reindex(index=newIndex)
    return df


def parseScore(f, fmt=None, fixedOffset=FIXEDOFFSET, eventBased=False):
    """

    Parameters
    ----------
    f :
        
    fmt :
         (Default value = None)
    fixedOffset :
         (Default value = FIXEDOFFSET)
    eventBased :
         (Default value = False)

    Returns
    -------

    """

    # Step 0: Use music21 to parse the score
    s = m21Parse(f, fmt)

    # Step 1: Parse and produce a salami-sliced dataset
    df = _initialDataFrame(s, fmt)
    # Step 2: Turn salami-slice into fixed-duration steps
    if not eventBased:
        df = _reindexDataFrame(df, fixedOffset=fixedOffset)
    return df



from musiclang.library import *
from musiclang import Note
from fractions import Fraction as frac

from musiclang import Score

class Pattern:

    bar_duration = None
    nb_instruments = 8 # Maximum number of instruments
    drums = False

    def __call__(self, instruments=None, **kwargs):

        pattern_dict = self.action(instruments=instruments, **kwargs)
        if self.drums > 0:
            pattern_dict = self.get_drums(self.drums, pattern_dict)

        return pattern_dict


    def to_json(self):
        return {
            "bar_duration": self.bar_duration,
            "nb_instruments": self.nb_instruments,
            "pattern": str(NC(**self.action([f'v__{i}' for i in range(self.nb_instruments)])))
        }

    @classmethod
    def from_json(cls, json_dict: dict, instruments: list):
        pattern: Score = Score.from_str(json_dict['pattern'])
        nb_instruments = json_dict['nb_instruments']
        all_instruments = pattern.score.keys()
        instruments = [(f'v__{i}', instr) for i, instr in enumerate(instruments[:nb_instruments])]
        pattern_dict = {new_instrument: pattern.score[base_instrument] for base_instrument, new_instrument in instruments}
        for instr in all_instruments:
            if not instr.startswith('v__'):
                pattern_dict[instr] = pattern.score[instr]
        return pattern_dict

    def action(self, instruments=None, **kwargs):
        raise NotImplemented

    def get_drums(self, drums, pattern_dict):
        return pattern_dict



class PluckedChords(Pattern):

    def __init__(self, pluck_spacing=frac(1, 4), bar_duration=4):
        self.pluck_spacing = pluck_spacing
        self.bar_duration = bar_duration

    def action(self, instruments=None, **kwargs):
        pattern_dict = {}
        for i, instr in enumerate(instruments):
            template_note = Note('x', i, 0, 1)
            local_melody = []
            if i > 0:
                local_melody += [r.set_duration(self.pluck_spacing * i)]
            local_melody += [template_note.set_duration(self.pluck_spacing)]
            local_melody += [l.set_duration(self.bar_duration - (i + 1) * self.pluck_spacing)]

            local_melody = sum(local_melody, None)

            pattern_dict[instr] = local_melody

        return pattern_dict


class Waltz(Pattern):

    bar_duration = 3
    spacing = 1

    def action(self, instruments=None, **kwargs):
        pattern_dict = {}
        for i, instr in enumerate(instruments):
            template_note = Note('x', i, 0, 1)
            local_melody = []
            if i == 0:
                local_melody = template_note.set_duration(3 * self.spacing).accent + bd1.set_duration(3 * self.spacing)
            if i > 0:
                silence = r.set_duration(self.spacing)
                note = template_note.set_duration(self.spacing)
                local_melody = (silence.copy().accent + note.copy().accent + note.copy().retarded) * 2

            pattern_dict[instr] = local_melody

        return pattern_dict


class Nocturne(Pattern):

    nb_instruments = 4
    bar_duration = 4
    spacing = frac(1)

    def action(self, instruments=None, **kwargs):
        pattern_dict = {}
        for i, instr in enumerate(instruments):
            if i == 0:
                notes = [Note('x', i, 0, 1).set_duration(self.spacing) for i in range(self.nb_instruments)]
                local_melody = notes[0] + notes[1] + notes[3] + notes[2]
            else:
                local_melody = r.set_duration(self.bar_duration)
            pattern_dict[instr] = local_melody

        return pattern_dict


class AlbertiBass(Pattern):
    nb_instruments = 4
    bar_duration = 4
    spacing = frac(1)

    def action(self, instruments=None, **kwargs):
        pattern_dict = {}
        for i, instr in enumerate(instruments):
            if i == 0:
                notes = [Note('x', i, 0, 1).set_duration(self.spacing) for i in range(self.nb_instruments)]
                local_melody = notes[0] + notes[2] + notes[1] + notes[2]
            else:
                local_melody = r.set_duration(self.bar_duration)
            pattern_dict[instr] = local_melody

        return pattern_dict

class StandardPlucked(Pattern):

    def __init__(self, bar_duration=4):
        self.bar_duration = bar_duration

    def action(self, instruments=None, **kwargs):
        pattern_dict = {}
        for i, instr in enumerate(instruments):
            template_note = Note('x', i, 0, 1)
            local_melody = template_note.set_duration(self.bar_duration).to_melody()
            pattern_dict[instr] = local_melody

        return pattern_dict


class EpicMusicTernary(Pattern):

    bar_duration = 6
    spacing = frac(1 / 2)

    def __init__(self, drums=False):
        self.drums = drums

    def action(self, instruments=None, **kwargs):
        pattern_dict = {}
        for i, instr in enumerate(instruments):
            template_note = Note('x', i, 0, 1)
            local_melody = []
            if i == 0:
                local_melody = template_note.set_duration(self.bar_duration).accent
            if i > 0:
                first = template_note.set_duration(2 * self.spacing)
                local_melody = first.accent + (bu1 + bd1 + bu0 + bu1).e + bu0 + (bu1 + bd1 + bu0 + bu1).e

            pattern_dict[instr] = local_melody
        if self.drums:
            pattern_dict = self.get_drums(self.drums, pattern_dict)
        return pattern_dict

    def get_drums(self, drums, pattern_dict):
        """
        Return a drum pattern associated with this
        Returns
        -------

        """

        pattern_dict['drums__0'] = crash.set_duration(self.bar_duration).accent
        pattern_dict['drums__1'] = (bd.set_duration(self.bar_duration//2).accent) + (sn.set_duration(self.bar_duration//2).accent)
        pattern_dict['drums__3'] = (hh.set_duration(3 * self.spacing) + (hh.set_duration(3 * self.spacing))) * 2
        if drums == 2:
            pattern_dict['drums__4'] = bd.set_duration(self.spacing/2) * int(2 * self.bar_duration / self.spacing)
            pattern_dict['drums__5'] = (ride.set_duration(3 * self.spacing) + (ride.set_duration(3 * self.spacing))) * 2

        return pattern_dict




from .element import Formatter

class TonalityElement(Formatter):

    def __init__(self, tone):
        self.tone = tone

    def to_text(self):
        return ' ' + str(self)

    def __repr__(self):
        return f'{self.tone}:'


from musiclang import Score
from musiclang.transform.melody.continuation import ContinuationWhenSameNote


class Composer:
    """
    Class for composing a piece automatically using an annotation generator model from musiclang_predict
    """

    def __init__(self, progressions, form, form_orchestra, orchestras, signature=(4, 4), continuation=True, **kwargs):
        self.form = form
        self.form_orchestra = form_orchestra
        self.signature = signature
        self.orchestras = orchestras
        self.progressions = progressions
        self.continuation = continuation

    def compose(self, **kwargs):
        """
        Main method to compose

        Parameters
        ----------
        kwargs

        Returns
        -------
        score: Score
        text: str
            Final annotation predicted by the auto composer

        """
        text = self.assemble_progressions(self.progressions)
        score = Score.from_annotation(text)
        # Use standard pipeline
        if self.continuation:
            score = ContinuationWhenSameNote()(score)
        return score, text

    def assemble_progressions(self, progressions):
        full_text = []
        for i, idx_form in enumerate(self.form):
            new_text = progressions[idx_form]
            full_text.append(self.orchestras[self.form_orchestra[i]])
            full_text.append(new_text)
        full_text = '\n'.join(full_text)
        full_text = self.clean(full_text)
        return full_text

    def line_is_bar(self, line):
        return line.lstrip(' ').startswith('m')

    def clean(self, progression):
        # Rename bar number to fit a progression
        # Basic replacements
        already_has_time_signature = False
        progression = progression.replace('iV', 'IV').replace('vI', 'VI').replace('iI', 'II')
        progression = progression.replace('Vi', 'VI')
        progression = progression.replace(':||', '').replace('||:', '')
        progression = progression.replace('||', '')
        new_lines = []
        for line in progression.split('\n'):
            line = line.lstrip('\t').lstrip(' ')
            if line.startswith('m') and '=' not in line:
                try:
                    elements = line.split(' ')
                    elements[0] = f'mx'
                    new_lines.append(' '.join(elements))
                except:
                    # FIXME : Repetitions are removed for the time being
                    pass

            elif line.capitalize().startswith('Time'):
                if not already_has_time_signature:
                    new_lines.append(line)
                    already_has_time_signature = True
            elif line.startswith('!'):
                new_lines.append(line)

        return '\n'.join(new_lines)


from .element import Formatter

class BarElement(Formatter):

    def __init__(self, idx=None):
        self.idx = idx

    def to_text(self):
        return '\n' + str(self)

    def __repr__(self):
        real_idx = str(self.idx) if self.idx is not None else 'x'
        return f'm{real_idx}'


from musiclang.transform import ChordTransformer


class Modulate(ChordTransformer):
    """
    Modulate to a different tonality
    """
    def __init__(self, modulation):
        self.modulation = modulation

    def action(self, chord, **kwargs):
        return chord % self.modulation


class ModulateKeepOriginMode(ChordTransformer):
    """
    Modulate to a different tonality, but each chord keeps the same tonality mode
    """
    def __init__(self, modulation):
        from musiclang import Tonality, Element
        if isinstance(modulation, int):
            self.modulation = Tonality(modulation)
        elif isinstance(modulation, Element):
            self.modulation = Tonality(modulation.val)
        else:
            self.modulation = modulation

    def action(self, chord, **kwargs):
        from musiclang import Tonality
        mode = chord.tonality.mode
        new_chord = (chord % self.modulation) % Tonality(0, mode=mode)
        return new_chord


class RepeatChord(ChordTransformer):
    """
    Repeat a chord n times
    """
    def __init__(self, n):
        self.n = n

    def action(self, chord, **kwargs):
        return chord * self.n



from fractions import Fraction as frac
import re
from musiclang.library import *

class BaseElement:

    def to_text(self):
        pass


class Bar(BaseElement):

    def __init__(self, idx=None):
        self.idx = idx

    def to_text(self):
        real_idx = str(self.idx) if self.idx is not None else 'x'
        return f'\nm{real_idx}'


class TonalityLine(BaseElement):

    def __init__(self, text):
        self.text = text
        self.tonality = None
        self.init()

    def init(self):
        self.tonality = self.text.split(':')[1].replace(' ', '')
    def to_text(self):
        return f'{self.key}:{self.mode}'

    def parse(self, score, parent):
        return CurrentTonality(self.tonality).parse(score, parent)
class CurrentTonality(BaseElement):

    def __init__(self, text):
        self.text = text
        self.key = 0
        self.mode = 'M'
        self.init()

    def to_text(self):
        return self.text + ' '

    def init(self):
        tab = 'CDEFGAB'
        notes = [0, 2, 4, 5, 7, 9, 11]
        note = self.text.replace(':', '').replace('#', '').replace('b', '').replace('-', '')
        tone = notes[tab.index(note.upper())]
        mode = 'major' if note.upper() == note else 'minor'

        tone += self.text.count('#')
        tone -= self.text.count('b')
        tone -= self.text.count('-')

        self.key = tone
        self.mode = mode

    def parse(self, score, parent):
        parent.key = self.key
        parent.mode = self.mode
        return score

    def __repr__(self):
        return f"CurrentTonality({self.key}, {self.mode})"



class BarLine:
    def __init__(self, text, parent):
        self.text = text
        self.idx = 0
        self.init(parent)

    def is_chord(self, element):
        return not self.is_beat(element)

    def is_beat(self, element):
        return element.startswith('b') and element[1].isdigit()

    def is_tonality(self, element):
        return ':' in element and not ('|' in element)

    def init(self, parent):
        # Get bar idx, ignoring var
        bar_number = self.text.split(' ')[0].split('var')[0]
        if 'x' in bar_number:
            self.idx = parent.init_bar_number + 1
        else:
            self.idx = int(bar_number[1:])

    def get_elements(self):
        array = self.text.split(' ')
        elements = []
        last_was_chord = False
        for element in array[1:]:
            if self.is_beat(element):
                elements.append(Beat(element))
                last_was_chord = False  # Only a beat can allow a new chord
            elif self.is_tonality(element):
                elements.append(CurrentTonality(element))
            elif self.is_chord(element) and not last_was_chord:
                if element != '':
                    elements.append(BarChord(element))
                    last_was_chord = True
        return elements

    def parse(self, score, parent):
        parent.set_bar_number(self.idx)
        return score

    def __repr__(self):
        return f"Bar({self.idx})"



class MultiBar:
    def __init__(self, text):
        self.text = text
        self.receiver_range = None
        self.emitter_range = None
        self.init()

    def is_chord(self, element):
        pass

    def is_beat(self, element):
        pass

    def init(self):
        new, old = self.text.replace(' ', '').split('=')
        res = new[1:].split('-')
        if len(res) == 1:
            a = res[0]
            b = a
        else:
            a, b = res

        a, b = int(a), int(b) + 1
        self.receiver_range = a, b

        res = old[1:].split('-')
        if len(res) == 1:
            c = res[0]
            d = c
        else:
            c, d = res
        c, d = int(c), int(d) + 1
        self.emitter_range = c, d
        if d - c != b - a:
            raise Exception(f'Range of bars must be equals for: {self.text}')

    def parse(self, score, parent):
        return score

    def __repr__(self):
        return f"MultiBar({self.text})"


class Event:

    def __new__(cls, text, parent):
        return cls.init(text, parent)

    @classmethod
    def is_generate(cls, text):
        return text.replace(' ', '').startswith('!generate')

    @classmethod
    def is_voicing(cls, text):
        return text.replace(' ', '').startswith('!voicing')

    @classmethod
    def is_voice_leading(cls, text):
        return text.replace(' ', '').startswith('!voice_leading')

    @classmethod
    def is_rhythm(cls, text):
        return text.replace(' ', '').startswith('!rhythm')

    @classmethod
    def is_instruments(cls, text):
        return text.replace(' ', '').startswith('!instruments')

    @classmethod
    def is_counterpoint(cls, text):
        return text.replace(' ', '').startswith('!counterpoint')

    @classmethod
    def init(cls, text, parent):
        if cls.is_generate(text):
            return Generate(text)
        elif cls.is_voicing(text):
            return Voicing(text)
        elif cls.is_instruments(text):
            return Instruments(text)
        elif cls.is_rhythm(text):
            return Rhythm(text, parent)
        elif cls.is_voice_leading(text):
            return VoiceLeading(text, parent)
        elif cls.is_counterpoint(text):
            return Counterpoint(text, parent)
        else:
            raise Exception(f'Not existing event {text}')




class Counterpoint:
    def __init__(self, text, parent):
        self.text = text
        self.init(parent)

    def init(self, parent):
        self.text = re.sub(' +', ' ', self.text)
        insts = ','.join([f"'{ins}'" for ins in self.text.replace('!counterpoint', '').lstrip(' ').split(' ')])
        insts = eval(insts + ',')
        tracks = {}
        # Assign tracks
        fixed_voices = []
        for ins in insts:
            if ins != '':
                if '__' in ins:
                    fixed_voices.append(ins)
                    tracks[ins] = int(ins.split('__')[1]) + 1
                else:
                    fixed_voices.append(f'{ins}__{tracks.get(ins, 0)}')
                    tracks[ins] = tracks.get(ins, 0) + 1

        self.fixed_voices = fixed_voices
        parent.set_counterpoint(self.fixed_voices)

    def parse(self, score, parent):

        return score

    def __repr__(self):
        return f"Counterpoint({self.text})"


class VoiceLeading:
    def __init__(self, text, parent):
        self.text = text
        self.init(parent)

    def init(self, parent):
        from musiclang.transform import VoiceLeading
        self.text = re.sub(' +', ' ', self.text)
        insts = ','.join([f"'{ins}'" for ins in self.text.replace('!voice_leading', '').lstrip(' ').split(' ')])
        insts = eval(insts + ',')
        tracks = {}
        # Assign tracks
        fixed_voices = []
        for ins in insts:
            if ins != '':
                if '__' in ins:
                    fixed_voices.append(ins)
                    tracks[ins] = int(ins.split('__')[1]) + 1
                else:
                    fixed_voices.append(f'{ins}__{tracks.get(ins, 0)}')
                    tracks[ins] = tracks.get(ins, 0) + 1
        self.voice_leading = VoiceLeading(fixed_voices=fixed_voices)
        parent.set_voice_leading(self.voice_leading)

    def parse(self, score, parent):

        return score

    def __repr__(self):
        return f"VoiceLeading({self.text})"


class Rhythm:
    """
    Apply a rhythm to a specific part
    """
    def __init__(self, text, parent):
        self.text = text
        self.init(parent)

    def init(self, parent):
        self.text = re.sub(' +', ' ', self.text)
        self.text = self.text.lstrip(' ')
        args = self.text.split(' ')
        self.instrument = f"{args[1]}"
        self.rhythm = f"{args[2]}"
        self.nb_bars = len(self.rhythm.split('|'))
        self.rhythm = [r if r in ['x', '.'] else '' for r in self.rhythm.replace('|', '')]
        self.rhythm = [1 if r == 'x' else 0 for r in self.rhythm]
        if '__' not in self.instrument:
            self.instrument = f"{self.instrument}__0"
        parent.set_rhythm(self.instrument, self.nb_bars, self.rhythm)

    def parse(self, score, parent):
        return score

    def __repr__(self):
        return f"Rhythm({self.text})"


class Instruments:
    def __init__(self, text):
        self.text = text
        self.init()

    def init(self):
        self.text = re.sub(' +', ' ', self.text)
        insts = ','.join([f"'{ins}'" for ins in self.text.replace('!instruments', '').lstrip(' ').split(' ')])
        insts = eval(insts)
        if isinstance(insts, str):
            insts = (insts,)
        tracks = {}
        # Assign tracks
        self.instruments = []
        for ins in insts:
            self.instruments.append(f'{ins}__{tracks.get(ins, 0)}')
            tracks[ins] = tracks.get(ins, 0) + 1

    def parse(self, score, parent):
        parent.instruments = self.instruments
        return score

    def __repr__(self):
        return f"Instruments({self.text})"

class Voicing:
    def __init__(self, text):
        self.text = text
        self.init()

    def init(self):
        import re
        self.text = re.sub(' +', ' ', self.text)
        self.score = ','.join([f"{ins}" for ins in self.text.replace('!voicing', '').lstrip(' ').split(' ')])
        self.score = eval(self.score)

    def parse(self, score, parent):
        parent.current_score = self.score
        return score

    def __repr__(self):
        return f"Voicing({self.text})"

class Generate:
    def __init__(self, text):
        self.text = text

    def parse(self, score, parent):
        return score

    def __repr__(self):
        return f"Generate({self.text})"


class TimeSignature:

    def __init__(self, text):
        self.text = text
        self.num = 4
        self.den = 4


    def parse(self, score, parent):
        num, den = self.text.split(':')[1].replace(' ', '').split('/')
        parent.set_time_signature((int(num), int(den)))
        return score

    def __repr__(self):
        return f'TimeSignature({self.num}, {self.den})'


class Beat:
    def __init__(self, text):
        self.text = text
        self.value = None
        self.init()

    def init(self):
        from fractions import Fraction as frac
        self.value = self.text.replace('b', '')

    def get_real_value(self, parent):
        CONVENTION_DICT = {(6, 8): 2, (2, 2): 2}
        nb_beats = CONVENTION_DICT.get(parent.time_signature, parent.duration)
        ratio = frac(nb_beats / parent.duration).limit_denominator(8)
        real_value = frac(float(self.value)).limit_denominator(8) - 1
        real_value /= ratio
        return real_value

    def parse(self, score, parent):
        parent.set_current_beat(self.get_real_value(parent))
        return score

    def __repr__(self):
        return f'Beat({self.value})'


class BarChord:

    def __init__(self, text):
        self.text = text

    def parse(self, score, parent):
        try:
            from musiclang import Chord, Tonality
            from musiclang.analyze.roman_parser import analyze_one_chord
            key, mode = parent.key, parent.mode
            final_degree, extension, final_key, final_mode = analyze_one_chord(self.text, key, mode)
            chord = Chord(final_degree,
                          tonality=Tonality(final_key, final_mode))[extension]
            new_dur = parent.duration - parent.current_beat
            applied_score = {key: item for key, item in zip(parent.instruments, parent.current_score)}
            chord = chord(**applied_score).set_duration(new_dur)
            score = parent.add_chord(chord, score)
        except Exception as e:
            new_exc = Exception(f"Exception for '{self.text}' occured at beat {parent.current_beat} of bar {parent.bar_number} : {e} ")
            print(new_exc)
            #raise new_exc from e
        return score

    def __repr__(self):
        return f"Chord({self.text})"


